"""Settings file for DJ Press."""

# DJPress settings
DJPRESS_SETTINGS = {
    "ARCHIVE_ENABLED": (True, bool),
    "ARCHIVE_PREFIX": ("", str),
    "AUTHOR_ENABLED": (True, bool),
    "AUTHOR_PREFIX": ("author", str),
    "BLOG_DESCRIPTION": ("", str),
    "BLOG_TITLE": ("My DJ Press Blog", str),
    "CACHE_CATEGORIES": (True, bool),
    "CACHE_RECENT_PUBLISHED_POSTS": (False, bool),
    "CATEGORY_ENABLED": (True, bool),
    "CATEGORY_PREFIX": ("category", str),
    "MARKDOWN_EXTENSION_CONFIGS": ({}, dict),
    "MARKDOWN_EXTENSIONS": ([], list),
    "MARKDOWN_RENDERER": ("djpress.markdown_renderer.default_renderer", str),
    "MICROFORMATS_ENABLED": (True, bool),
    "PLUGINS": ([], list),
    "PLUGIN_SETTINGS": ({}, dict),
    "POST_PREFIX": ("{{ year }}/{{ month }}/{{ day }}", str),
    "POST_READ_MORE_TEXT": ("Read more...", str),
    "RECENT_PUBLISHED_POSTS_COUNT": (20, int),
    "RSS_ENABLED": (True, bool),
    "RSS_PATH": ("rss", str),
    "THEME": ("default", str),
    "TRUNCATE_TAG": ("<!--more-->", str),
}



"""Post model."""

import logging

from django.contrib.auth.models import User
from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Max
from django.utils import timezone
from django.utils.text import slugify

from djpress.conf import settings as djpress_settings
from djpress.exceptions import PageNotFoundError, PostNotFoundError
from djpress.models import Category
from djpress.plugins import Hooks, registry
from djpress.utils import get_markdown_renderer

logger = logging.getLogger(__name__)

render_markdown = get_markdown_renderer()


PUBLISHED_POSTS_CACHE_KEY = "published_posts"


class AdminManager(models.Manager):
    """Manager that returns all posts/pages - used only by admin."""

    def get_queryset(self) -> models.QuerySet:
        """Return completely unfiltered queryset."""
        return super().get_queryset()


class PostsAndPagesManager(models.Manager):
    """Default manager that only returns published content."""

    def get_queryset(self) -> models.QuerySet:
        """Return only published posts and pages."""
        return (
            super()
            .get_queryset()
            .filter(
                status="published",
                date__lte=timezone.now(),
            )
        )


class PagesManager(models.Manager):
    """Page custom manager."""

    def get_queryset(self: "PagesManager") -> models.QuerySet:
        """Return the queryset for pages."""
        return (
            super()
            .get_queryset()
            .filter(
                post_type="page",
                status="published",
                date__lte=timezone.now(),
            )
            .order_by("menu_order", "title")
        )

    def get_published_pages(self: "PagesManager") -> models.QuerySet:
        """Return all published pages.

        For a page to be considered published, it must meet the following requirements:
        - The status must be "published".
        - The date must be less than or equal to the current date/time.
        - All parent pages must also be published.
        """
        return Post.page_objects.filter(
            pk__in=[page.pk for page in self.get_queryset() if page.is_published],
        ).order_by("menu_order", "title", "-date")

    def get_published_page_by_slug(
        self: "PagesManager",
        slug: str,
    ) -> "Post":
        """Return a single published page.

        Args:
            slug (str): The slug of the page.

        Returns:
            Post: The published page.
        """
        try:
            page: Post = self.get_published_pages().get(slug=slug)
        except Post.DoesNotExist as exc:
            msg = "Page not found"
            raise PageNotFoundError(msg) from exc

        return page

    def get_published_page_by_path(
        self: "PagesManager",
        path: str,
    ) -> "Post":
        """Return a single published page from a path.

        The path can consist of one or more pages, e.g. "about", "about/contact". We split the path into parts and
        check the page is valid. These are the checks to be made:

        1. The first part of the path must be a top-level page.
        2. Subsequent parts must be children of the previous page.
        3. The final page must be published.

        Args:
            path (str): The path to the page.

        Returns:
            Post: The published page.

        Raises:
            PageNotFoundError: If the page cannot be found.
        """
        # Strip leading and trailing slashes and split the path into parts
        path_parts = path.strip("/").split("/")

        current_page = None

        for i, slug in enumerate(path_parts):
            if i == 0:
                try:
                    # The first item must be a top-level page
                    current_page = self.get(slug=slug, parent__isnull=True)
                except Post.DoesNotExist as exc:
                    msg = "Page not found"
                    raise PageNotFoundError(msg) from exc
            else:
                try:
                    # Subsequent items must be children of the previous page
                    current_page = self.get(slug=slug, parent=current_page)
                except Post.DoesNotExist as exc:
                    msg = "Page not found"
                    raise PageNotFoundError(msg) from exc

        return current_page

    def get_page_tree(self) -> list[dict["Post", list[dict]]]:
        """Return the page tree.

        This returns a list of top-level pages. Each page is a dict containing the Post object and a list of children.

        Used to build the page hierarchy.

        ```
        root_pages = [
            {
                'page': <Page object>,
                'children': [
                    {
                        'page': <Child Page object>,
                        'children': [
                            {
                                'page': <Grandchild Page object>,
                                'children': []
                            },
                            # ... more grandchildren ...
                        ]
                    },
                    # ... more children ...
                ]
            },
            # ... more root pages ...
        ]
        ```

        Returns:
            list[dict["Post", list[dict]]]: A list of top-level pages - each page is a dict containing the Post object
            and a list of children. Each child is a dict containing the Post object and a list of children, and so on.
        """
        pages = self.get_published_pages().select_related("parent")
        page_dict = {page.id: {"page": page, "children": []} for page in pages}
        root_pages = []
        for page_data in page_dict.values():
            page = page_data["page"]
            if page.parent:
                page_dict[page.parent.id]["children"].append(page_data)

            else:
                root_pages.append(page_data)
        return root_pages


class PostsManager(models.Manager):
    """Post custom manager."""

    def get_queryset(self: "PostsManager") -> models.QuerySet:
        """Return the queryset for posts."""
        return (
            super()
            .get_queryset()
            .filter(
                post_type="post",
                status="published",
                date__lte=timezone.now(),
            )
            .order_by("-date")
        )

    def get_published_posts(self: "PostsManager") -> models.QuerySet:
        """Returns all published posts.

        For a post to be considered published, it must meet the following requirements:
        - The status must be "published".
        - The date must be less than or equal to the current date/time.
        """
        return self.get_queryset().prefetch_related("categories", "author")

    def get_recent_published_posts(self: "PostsManager") -> models.QuerySet:
        """Return recent published posts.

        This does not return a paginated queryset. Use get_paginated_published_posts
        instead.

        If CACHE_RECENT_PUBLISHED_POSTS is set to True, we return the cached queryset.
        """
        if djpress_settings.CACHE_RECENT_PUBLISHED_POSTS:
            return self._get_cached_recent_published_posts()

        return self.get_published_posts()[: djpress_settings.RECENT_PUBLISHED_POSTS_COUNT]

    def _get_cached_recent_published_posts(self: "PostsManager") -> models.QuerySet:
        """Return the cached recent published posts queryset.

        If there are any future posts, we calculate the seconds until that post, then we
        set the timeout to that number of seconds.
        """
        queryset = cache.get(PUBLISHED_POSTS_CACHE_KEY)

        # Check if the cache is empty or if the length of the queryset is not equal to the number of recent posts. If
        # the length is different it means the setting may have changed.
        if queryset is None or len(queryset) != djpress_settings.RECENT_PUBLISHED_POSTS_COUNT:
            # Get the queryset from the database for all published posts, including those in the future. Then we
            # calculate the timeout to set, and then filter the queryset to only include the recent published posts.
            # Note: we use admin_objects here to get all posts, including those in the future.
            queryset = (
                self.model.admin_objects.filter(post_type="post", status="published")
                .prefetch_related(
                    "categories",
                    "author",
                )
                .order_by("-date")
            )
            timeout = self._get_cache_timeout(queryset)
            queryset = queryset.filter(date__lte=timezone.now())[: djpress_settings.RECENT_PUBLISHED_POSTS_COUNT]

            cache.set(
                PUBLISHED_POSTS_CACHE_KEY,
                queryset,
                timeout=timeout,
            )

        return queryset

    def _get_cache_timeout(
        self: "PostsManager",
        queryset: models.QuerySet,
    ) -> int | None:
        """Return the timeout for the cache.

        If there are any future posts, we calculate the seconds until that post.

        Args:
            queryset: The queryset of published posts.

        Returns:
            int | None: The number of seconds until the next future post, or None if
            there are no future posts.
        """
        # TODO: total_seconds returns a float not an int.
        future_posts = queryset.filter(date__gt=timezone.now())
        if future_posts.exists():
            future_post = future_posts[0]
            return int((future_post.date - timezone.now()).total_seconds())

        return None

    def get_published_post_by_slug(
        self: "PostsManager",
        slug: str,
        year: int | None = None,
        month: int | None = None,
        day: int | None = None,
    ) -> "Post":
        """Return a single published post.

        Args:
            slug (str): The post slug.
            year (int | None): The year.
            month (int | None): The month.
            day (int | None): The day.

        Returns:
            Post: The published post.

        Raises:
            PostNotFoundError: If the post is not found in the database.
        """
        # TODO: try to get the post from the cache

        filters = {"slug": slug}

        if year:
            filters["date__year"] = year

        if month:
            filters["date__month"] = month

        if day:
            filters["date__day"] = day

        try:
            return self.get_published_posts().get(**filters)
        except Post.DoesNotExist as exc:
            msg = "Post not found"
            raise PostNotFoundError(msg) from exc

    def get_published_posts_by_category(
        self: "PostsManager",
        category: Category,
    ) -> models.QuerySet:
        """Return all published posts for a given category.

        Must have a date less than or equal to the current date/time for a specific
        category, ordered by date in descending order.
        """
        return self.get_published_posts().filter(categories=category)

    def get_published_posts_by_author(
        self: "PostsManager",
        author: User,
    ) -> models.QuerySet:
        """Return all published posts for a given author.

        Must have a date less than or equal to the current date/time for a specific
        author, ordered by date in descending order.
        """
        return self.get_published_posts().filter(author=author)

    def get_years(self) -> models.QuerySet:
        """Return a list of years that have published posts.

        Returns:
            list[int]: A distinct list of years.
        """
        return self.dates("date", "year")

    def get_months(self, year: int) -> models.QuerySet:
        """Return a list of months for a given year that have published posts.

        Args:
            year (int): The year.

        Returns:
            list[int]: A distinct list of months.
        """
        return self.filter(date__year=year).dates("date", "month")

    def get_days(self, year: int, month: int) -> models.QuerySet:
        """Return a list of days for a given year and month that have published posts.

        Args:
            year (int): The year.
            month (int): The month.

        Returns:
            list[int]: A distinct list of days.
        """
        return self.filter(date__year=year, date__month=month).dates("date", "day")

    def get_year_last_modified(self, year: int) -> timezone.datetime | None:
        """Return the most recent modified_date of posts for a given year.

        Args:
            year (int): The year.

        Returns:
            timezone.datetime | None: The last published post for the given year.
        """
        return self.filter(date__year=year).aggregate(latest=Max("modified_date"))["latest"]

    def get_month_last_modified(self, year: int, month: int) -> timezone.datetime | None:
        """Return the most recent modified_date of posts for a given month.

        Args:
            year (int): The year.
            month (int): The month.

        Returns:
            timezone.datetime | None: The last published post for the given month.
        """
        return self.filter(date__year=year, date__month=month).aggregate(latest=Max("modified_date"))["latest"]

    def get_day_last_modified(self, year: int, month: int, day: int) -> timezone.datetime | None:
        """Return the most recent modified_date of posts for a given day.

        Args:
            year (int): The year.
            month (int): The month.
            day (int): The day.

        Returns:
            timezone.datetime | None: The last published post for the given day.
        """
        return self.filter(date__year=year, date__month=month, date__day=day).aggregate(latest=Max("modified_date"))[
            "latest"
        ]


class Post(models.Model):
    """Post model."""

    STATUS_CHOICES = [("draft", "Draft"), ("published", "Published")]
    CONTENT_TYPE_CHOICES = [("post", "Post"), ("page", "Page")]

    title = models.CharField(max_length=200)
    slug = models.SlugField(unique=True, blank=True)
    content = models.TextField()
    author = models.ForeignKey(User, on_delete=models.CASCADE)
    date = models.DateTimeField(default=timezone.now)
    modified_date = models.DateTimeField(auto_now=True)
    status = models.CharField(max_length=10, choices=STATUS_CHOICES, default="draft")
    post_type = models.CharField(max_length=10, choices=CONTENT_TYPE_CHOICES, default="post")
    categories = models.ManyToManyField(Category, blank=True, related_name="_posts")
    menu_order = models.IntegerField(default=0)
    parent = models.ForeignKey(
        "self",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="_children",  # Danger! This returns all children of a parent, published or not.
        limit_choices_to={"post_type": "page"},
    )

    # Managers
    admin_objects = AdminManager()  # Unfiltered - for admin use only
    objects = PostsAndPagesManager()  # Default manager returns only published content
    page_objects: "PagesManager" = PagesManager()
    post_objects: "PostsManager" = PostsManager()

    class Meta:
        """Meta options for the Post model."""

        verbose_name = "post"
        verbose_name_plural = "posts"
        default_manager_name = "admin_objects"  # For Django internals/admin

    def __str__(self: "Post") -> str:
        """Return the string representation of the post."""
        return self.title

    def save(self: "Post", *args, **kwargs) -> None:  # noqa: ANN002, ANN003
        """Override the save method."""
        # auto-generate the slug.
        if not self.slug:
            self.slug = slugify(self.title)
            if not self.slug or self.slug.strip("-") == "":
                msg = "Invalid title. Unable to generate a valid slug."
                raise ValueError(msg)

        # If the post is a post, we need to ensure that the parent is None
        if self.post_type == "post":
            self.parent = None

        self.full_clean()
        super().save(*args, **kwargs)

        # If the post is a post and it's published, run the post_save_post hook
        if self.post_type == "post" and self.is_published:
            registry.run_hook(Hooks.POST_SAVE_POST, self)

    def clean(self) -> None:
        """Custom validation for the Post model."""
        # Check for circular references in the page hierarchy
        self._check_circular_reference()

    def _check_circular_reference(self) -> None:
        """Check for circular references in the page hierarchy.

        This is a recursive function that checks if the current page is an ancestor of itself. This is needed to ensure
        that we don't create a circular reference in the page hierarchy. This is called in the clean method.

        For example, we need to avoid the following page hierarchy from happening:
        - Page A
            - Page B
                - Page C
                    - Page A

        Returns:
            None

        Raises:
            ValidationError: If a circular reference is detected.
        """
        # If there's no parent, we don't need to check for circular references
        if not self.parent:
            return

        ancestor = self.parent
        while ancestor:
            if ancestor.pk == self.pk:
                msg = "Circular reference detected in page hierarchy."
                raise ValidationError(msg)
            ancestor = ancestor.parent

    @property
    def children(self) -> models.QuerySet:
        """Return only published children pages."""
        return self._children.filter(
            status="published",
            date__lte=timezone.now(),
        ).order_by("menu_order", "title")

    @property
    def content_markdown(self: "Post") -> str:
        """Return the content as HTML converted from Markdown."""
        # Get the raw markdown content
        content = self.content

        # Let plugins modify the markdown before rendering
        content = registry.run_hook(Hooks.PRE_RENDER_CONTENT, content)

        # Render the markdown
        html_content = render_markdown(content)

        # Let the plugins modify the markdown after rendering and return the results
        return registry.run_hook(Hooks.POST_RENDER_CONTENT, html_content)

    @property
    def truncated_content_markdown(self: "Post") -> str:
        """Return the truncated content as HTML converted from Markdown."""
        read_more_index = self.content.find(djpress_settings.TRUNCATE_TAG)
        truncated_content = self.content[:read_more_index] if read_more_index != -1 else self.content
        return render_markdown(truncated_content)

    @property
    def is_truncated(self: "Post") -> bool:
        """Return whether the content is truncated."""
        return djpress_settings.TRUNCATE_TAG in self.content

    @property
    def url(self: "Post") -> str:
        """Return the post's URL.

        Returns:
            str: The post's URL.
        """
        from djpress.url_utils import get_page_url, get_post_url

        if self.post_type == "page":
            return get_page_url(self)

        return get_post_url(self)

    @property
    def permalink(self: "Post") -> str:
        """Return the post's permalink.

        The posts permalink is constructed of the following elements:
        - The post prefix - this is configured in POST_PREFIX and could be an empty
          string or a custom string.
        - The post slug - this is a unique identifier for the post. TODO: should this be
          a database unique constraint, or should we handle it in software instead?
        """
        # If the post type is a page, we return the full page path
        if self.post_type == "page":
            return self.full_page_path

        prefix = djpress_settings.POST_PREFIX

        # Replace placeholders in POST_PREFIX with actual values
        replacements = {
            "{{ year }}": self.date.strftime("%Y"),
            "{{ month }}": self.date.strftime("%m"),
            "{{ day }}": self.date.strftime("%d"),
        }

        for placeholder, value in replacements.items():
            prefix = prefix.replace(placeholder, value)

        # Ensure there's no leading or trailing slash, then join with the slug
        url_parts = [part for part in prefix.split("/") if part] + [self.slug]

        return "/".join(url_parts)

    @property
    def full_page_path(self) -> str:
        """Return the full page path.

        This is the full path to the page, including any parent pages.

        Returns:
            str: The full page path.
        """
        if self.parent:
            return f"{self.parent.full_page_path}/{self.slug}"
        return self.slug

    @property
    def is_published(self: "Post") -> bool:
        """Return whether the post or page is published.

        For a post to be published, it must meet the following requirements:
        - The status must be "published".
        - The date must be less than or equal to the current date/time.

        For a page to be published, it must meet the following requirements:
        - The status must be "published".
        - The date must be less than or equal to the current date/time.
        - All ancestor pages must also be published.

        Returns:
            bool: Whether the post is published.
        """
        # If the post or page status is not published or the date is in the future, return False
        if not (self.status == "published" and self.date <= timezone.now()):
            return False

        # If the post is a page and has a parent, check if the parent is published
        if self.post_type == "page" and self.parent:
            return self.parent.is_published

        # If we get to here, the post is published
        return True

    @property
    def is_parent(self) -> bool:
        """Return whether the post is a parent page.

        Returns:
            bool: Whether the post is a parent page.
        """
        return self.children.exists()

    @property
    def is_child(self) -> bool:
        """Return whether the post is a child page.

        Returns:
            bool: Whether the post is a child page.
        """
        return self.parent is not None



from collections.abc import Iterable

import pytest

from datetime import datetime

from django.contrib.auth.models import User
from django.urls import reverse
from django.utils import timezone

from djpress.models import Post
from djpress.url_utils import get_archives_url, get_author_url, get_category_url


@pytest.mark.django_db
def test_index_view_no_posts(client):
    url = reverse("djpress:index")
    response = client.get(url)
    assert response.status_code == 200
    assert b"No posts available" in response.content
    assert "posts" in response.context
    assert isinstance(response.context["posts"], Iterable)


@pytest.mark.django_db
def test_index_view_multiple_posts(client, test_post1, test_post2, test_post3):
    url = "/"
    response = client.get(url)
    assert response.status_code == 200
    assert "posts" in response.context
    assert isinstance(response.context["posts"], Iterable)
    assert (
        f'<h2 class="p-name"><a href="/test-posts/test-post1/" title="Test Post1" class="u-url">Test Post1</a></h2>'
        in response.content.decode()
    )
    assert f'<article class="h-entry">' in response.content.decode()


@pytest.mark.django_db
def test_single_post_view(client, test_post1):
    url = test_post1.url
    response = client.get(url)
    assert response.status_code == 200
    assert "post" in response.context
    assert not isinstance(response.context["post"], Iterable)
    assert f'<h1 class="p-name">{test_post1.title}</h1>' in response.content.decode()


@pytest.mark.django_db
def test_single_post_view_with_date(client, settings, test_post1):
    assert settings.DJPRESS_SETTINGS["POST_PREFIX"] == "test-posts"
    settings.DJPRESS_SETTINGS["POST_PREFIX"] = "{{ year }}/{{ month }}/{{ day }}"
    dt = datetime.now()
    year = dt.strftime("%Y")
    month = dt.strftime("%m")
    day = dt.strftime("%d")
    url = f"/{year}/{month}/{day}/{test_post1.slug}/"
    response = client.get(url)
    assert response.status_code == 200


@pytest.mark.django_db
def test_single_post_view_wrong_date(client, settings, test_post1):
    assert settings.DJPRESS_SETTINGS["POST_PREFIX"] == "test-posts"
    settings.DJPRESS_SETTINGS["POST_PREFIX"] = "{{ year }}/{{ month }}/{{ day }}"
    url = f"/1999/01/01/{test_post1.slug}/"
    response = client.get(url)
    assert response.status_code == 404


@pytest.mark.django_db
def test_single_post_not_exist(client, settings):
    assert settings.DJPRESS_SETTINGS["POST_PREFIX"] == "test-posts"
    url = "/test-posts/foobar-does-not-exist/"
    response = client.get(url)
    assert response.status_code == 404


@pytest.mark.django_db
def test_single_page_view(client, test_page1):
    url = test_page1.url
    response = client.get(url)
    assert response.status_code == 200
    assert "post" in response.context
    assert not isinstance(response.context["post"], Iterable)


@pytest.mark.django_db
def test_author_with_no_posts_view(client, user):
    url = get_author_url(user)
    print(url)
    response = client.get(url)
    assert response.status_code == 200
    assert "author" in response.context
    assert b"No posts available" in response.content
    assert "posts" in response.context
    assert isinstance(response.context["posts"], Iterable)


@pytest.mark.django_db
def test_author_with_posts_view(client, test_post1):
    url = get_author_url(test_post1.author)
    response = client.get(url)
    assert response.status_code == 200
    assert "author" in response.context
    assert b"No posts available" not in response.content
    assert "posts" in response.context
    assert isinstance(response.context["posts"], Iterable)


@pytest.mark.django_db
def test_author_with_invalid_author(client, settings):
    assert settings.DJPRESS_SETTINGS["AUTHOR_PREFIX"] == "test-url-author"
    url = "/test-url-author/non-existent-author/"
    response = client.get(url)
    assert response.status_code == 404


@pytest.mark.django_db
def test_author_with_author_enabled_false(client, settings):
    """This will try to get a page from the database that does not exist."""
    assert settings.DJPRESS_SETTINGS["AUTHOR_ENABLED"] == True
    settings.DJPRESS_SETTINGS["AUTHOR_ENABLED"] = False
    url = "/test-url-author/non-existent-author/"
    response = client.get(url)
    assert response.status_code == 404


@pytest.mark.django_db
def test_category_with_no_posts_view(client, category1):
    url = get_category_url(category1)
    response = client.get(url)
    assert response.status_code == 200
    assert "category" in response.context
    assert b"No posts available" in response.content
    assert "posts" in response.context
    assert isinstance(response.context["posts"], Iterable)


@pytest.mark.django_db
def test_category_with_posts_view(client, category1, test_post1):
    test_post1.categories.add(category1)
    url = get_category_url(category1)
    response = client.get(url)
    assert response.status_code == 200
    assert "category" in response.context
    assert b"No posts available" not in response.content
    assert "posts" in response.context
    assert isinstance(response.context["posts"], Iterable)


@pytest.mark.django_db
def test_category_with_invalid_category(client, settings):
    assert settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"] == "test-url-category"
    url = "/test-url-category/non-existent-category/"
    response = client.get(url)
    assert response.status_code == 404


@pytest.mark.django_db
def test_category_with_category_enabled_false(client, settings):
    """This will try to get a page from the database that does not exist."""
    assert settings.DJPRESS_SETTINGS["CATEGORY_ENABLED"] == True
    settings.DJPRESS_SETTINGS["CATEGORY_ENABLED"] = False
    url = "/test-url-category/non-existent-category/"
    response = client.get(url)
    assert response.status_code == 404


@pytest.mark.django_db
def test_date_archives_year(client, settings, test_post1):
    assert settings.DJPRESS_SETTINGS["ARCHIVE_PREFIX"] == "test-url-archives"
    url = get_archives_url(test_post1.date.year)
    assert url == f"/test-url-archives/{test_post1.date.year}/"
    response = client.get(url)
    assert response.status_code == 200
    assert test_post1.title.encode() in response.content
    assert "posts" in response.context
    assert isinstance(response.context["posts"], Iterable)
    assert "Test Post1" in response.content.decode()

    settings.DJPRESS_SETTINGS["ARCHIVE_PREFIX"] = ""
    url = get_archives_url(test_post1.date.year)
    assert url == f"/{test_post1.date.year}/"

    response = client.get(url)
    assert response.status_code == 200
    assert test_post1.title.encode() in response.content
    assert "posts" in response.context
    assert isinstance(response.context["posts"], Iterable)


@pytest.mark.django_db
def test_date_archives_year_invalid_year(client, settings):
    assert settings.DJPRESS_SETTINGS["ARCHIVE_PREFIX"] == "test-url-archives"
    url = "/test-url-archives/0000/"
    response = client.get(url)
    assert response.status_code == 400


@pytest.mark.django_db
def test_date_archives_year_no_posts(client, test_post1):
    url = get_archives_url(test_post1.date.year - 1)
    response = client.get(url)
    assert response.status_code == 200
    assert not test_post1.title.encode() in response.content
    assert b"No posts available" in response.content
    assert "posts" in response.context
    assert isinstance(response.context["posts"], Iterable)


@pytest.mark.django_db
def test_date_archives_month(client, settings, test_post1):
    assert settings.DJPRESS_SETTINGS["ARCHIVE_PREFIX"] == "test-url-archives"
    url = get_archives_url(test_post1.date.year, test_post1.date.month)
    assert url == f"/test-url-archives/{test_post1.date.year}/{test_post1.date.month}/"

    response = client.get(url)
    assert response.status_code == 200
    assert test_post1.title.encode() in response.content
    assert "posts" in response.context
    assert isinstance(response.context["posts"], Iterable)

    settings.DJPRESS_SETTINGS["ARCHIVE_PREFIX"] = ""
    url = get_archives_url(test_post1.date.year, test_post1.date.month)
    assert url == f"/{test_post1.date.year}/{test_post1.date.month}/"

    response = client.get(url)
    assert response.status_code == 200
    assert test_post1.title.encode() in response.content
    assert "posts" in response.context
    assert isinstance(response.context["posts"], Iterable)


@pytest.mark.django_db
def test_date_archives_month_invalid_month(client, settings):
    assert settings.DJPRESS_SETTINGS["ARCHIVE_PREFIX"] == "test-url-archives"
    response1 = client.get("/test-url-archives/2024/00/")
    assert response1.status_code == 400
    response2 = client.get("/test-url-archives/2024/13/")
    assert response2.status_code == 400


@pytest.mark.django_db
def test_date_archives_month_no_posts(client, test_post1):
    url = get_archives_url(test_post1.date.year - 1, test_post1.date.month)
    response = client.get(url)
    assert response.status_code == 200
    assert not test_post1.title.encode() in response.content
    assert b"No posts available" in response.content
    assert "posts" in response.context
    assert isinstance(response.context["posts"], Iterable)


@pytest.mark.django_db
def test_date_archives_day(client, settings, test_post1):
    assert settings.DJPRESS_SETTINGS["ARCHIVE_PREFIX"] == "test-url-archives"
    url = get_archives_url(test_post1.date.year, test_post1.date.month, test_post1.date.day)
    assert url == f"/test-url-archives/{test_post1.date.year}/{test_post1.date.month}/{test_post1.date.day}/"

    response = client.get(url)
    assert response.status_code == 200
    assert test_post1.title.encode() in response.content
    assert "posts" in response.context
    assert isinstance(response.context["posts"], Iterable)

    settings.DJPRESS_SETTINGS["ARCHIVE_PREFIX"] = ""
    url = get_archives_url(test_post1.date.year, test_post1.date.month, test_post1.date.day)
    assert url == f"/{test_post1.date.year}/{test_post1.date.month}/{test_post1.date.day}/"

    response = client.get(url)
    assert response.status_code == 200
    assert test_post1.title.encode() in response.content
    assert "posts" in response.context
    assert isinstance(response.context["posts"], Iterable)

    assert settings.DJPRESS_SETTINGS["POST_PREFIX"] == "test-posts"
    settings.DJPRESS_SETTINGS["POST_PREFIX"] = "{{ year }}/{{ month }}/{{ day }}"
    url = get_archives_url(test_post1.date.year, test_post1.date.month, test_post1.date.day)
    assert url == f"/{test_post1.date.year}/{test_post1.date.month}/{test_post1.date.day}/"

    response = client.get(url)
    assert response.status_code == 200
    assert test_post1.title.encode() in response.content
    assert "posts" in response.context
    assert isinstance(response.context["posts"], Iterable)


@pytest.mark.django_db
def test_conflict_day_archives_and_single_post(client, settings, test_post1):
    """Change the POST_PREFIX to match the same format as the day archives.

    For example, with POST_PREFIX = "{{ year }}/{{ month }}", either of the following two URLs could be valid posts:

    - /2024/01/31/ - This could be a day archive, or a post with the slug "31" in the year 2024 and month 01.
    - /2024/01/test-post1/

    The first URL will try to get a post and if that doesn't exist, it will try to get a day archive, but only if the
    POST_PREFIX matches the day archive format.
    """
    settings.DJPRESS_SETTINGS["ARCHIVE_PREFIX"] = ""
    settings.DJPRESS_SETTINGS["POST_PREFIX"] = "{{ year }}/{{ month }}"
    url = get_archives_url(test_post1.date.year, test_post1.date.month, test_post1.date.day)
    assert url == f"/{test_post1.date.year}/{test_post1.date.month}/{test_post1.date.day}/"

    response = client.get(url)
    assert response.status_code == 200
    assert test_post1.title.encode() in response.content
    assert "posts" in response.context
    assert isinstance(response.context["posts"], Iterable)


@pytest.mark.django_db
def test_conflict_month_archives_and_single_post(client, settings, test_post1):
    """Similar concept to test_conflict_day_archives_and_single_post.

    With POST_PREFIX = "{{ year }}", either of the following two URLs could be valid posts:

    - /2024/12/ - This could be a month archive, or a post with the slug "12" in the year 2024.
    - /2024/test-post1/
    """
    settings.DJPRESS_SETTINGS["ARCHIVE_PREFIX"] = ""
    settings.DJPRESS_SETTINGS["POST_PREFIX"] = "{{ year }}"
    url = get_archives_url(test_post1.date.year, test_post1.date.month)
    assert url == f"/{test_post1.date.year}/{test_post1.date.month}/"

    response = client.get(url)
    assert response.status_code == 200
    assert test_post1.title.encode() in response.content
    assert "posts" in response.context
    assert isinstance(response.context["posts"], Iterable)


@pytest.mark.django_db
def test_conflict_year_archives_and_single_post(client, settings, test_post1):
    """Similar concept to test_conflict_day_archives_and_single_post and test_conflict_month_archives_and_single_post.

    With POST_PREFIX = "", either of the following two URLs could be valid posts:

    - /2024/ - This could be a year archive, or a post with the slug "2024".
    - /test-post1/
    """
    settings.DJPRESS_SETTINGS["ARCHIVE_PREFIX"] = ""
    settings.DJPRESS_SETTINGS["POST_PREFIX"] = ""
    url = get_archives_url(test_post1.date.year)
    assert url == f"/{test_post1.date.year}/"

    response = client.get(url)
    assert response.status_code == 200
    assert test_post1.title.encode() in response.content
    assert "posts" in response.context
    assert isinstance(response.context["posts"], Iterable)

    # If the post slug is 2024, then the post will be returned.
    test_post1.slug = str(test_post1.date.year)
    test_post1.save()
    response = client.get(url)
    assert response.status_code == 200
    assert test_post1.title.encode() in response.content
    assert "post" in response.context
    assert isinstance(response.context["post"], Post)


@pytest.mark.django_db
def test_date_archives_day_invalid_day(client, settings):
    assert settings.DJPRESS_SETTINGS["ARCHIVE_PREFIX"] == "test-url-archives"
    response1 = client.get("/test-url-archives/2024/01/00/")
    assert response1.status_code == 400
    response2 = client.get("/test-url-archives/2024/01/32/")
    assert response2.status_code == 400


@pytest.mark.django_db
def test_date_archives_day_no_posts(client, test_post1):
    url = get_archives_url(test_post1.date.year - 1, test_post1.date.month, test_post1.date.day)
    response = client.get(url)
    assert response.status_code == 200
    assert not test_post1.title.encode() in response.content
    assert b"No posts available" in response.content
    assert "posts" in response.context
    assert isinstance(response.context["posts"], Iterable)



"""Helper functions for the template tags."""

from django import template
from django.db import models
from django.utils.safestring import mark_safe

from djpress.conf import settings as djpress_settings
from djpress.models import Category, Post


def categories_html(
    categories: models.QuerySet,
    outer: str,
    outer_class: str,
    link_class: str,
) -> str:
    """Return the HTML for the categories.

    Note this isn't a template tag, but a helper function for the other template tags

    Args:
        categories: The categories.
        outer: The outer HTML tag for the categories.
        outer_class: The CSS class(es) for the outer tag.
        link_class: The CSS class(es) for the link.

    Returns:
        str: The HTML for the categories.
    """
    output = ""

    outer_class_html = f' class="{outer_class}"' if outer_class else ""

    if outer == "ul":
        output += f"<ul{outer_class_html}>"
        for category in categories:
            output += f"<li>{category_link(category, link_class)}</li>"
        output += "</ul>"

    if outer == "div":
        output += f"<div{outer_class_html}>"
        for category in categories:
            output += f"{category_link(category, link_class)}, "
        output = output[:-2]  # Remove the trailing comma and space
        output += "</div>"

    if outer == "span":
        output += f"<span{outer_class_html}>"
        for category in categories:
            output += f"{category_link(category, link_class)}, "
        output = output[:-2]  # Remove the trailing comma and space
        output += "</span>"

    return output


def category_link(category: Category, link_class: str = "") -> str:
    """Return the category link.

    This is not intended to be used as a template tag. It is used by the other
    template tags in this module to generate the category links.

    Args:
        category: The category.
        link_class: The CSS class(es) for the link.
    """
    category_url = category.url

    link_classes = ""

    # Add p-category if microformats are enabled
    if djpress_settings.MICROFORMATS_ENABLED:
        link_classes += "p-category "

    # Add the user-defined link class
    link_classes += link_class

    # Trim any trailing spaces
    link_classes = link_classes.strip()

    link_class_html = f' class="{link_classes}"' if link_classes else ""

    return (
        f'<a href="{category_url}" title="View all posts in the {category.title} '
        f'category"{link_class_html}>{category.title}</a>'
    )


def get_page_link(page: Post, link_class: str = "") -> str:
    """Return the page link.

    This is not intended to be used as a template tag. It is used by the other
    template tags in this module to generate the page links.

    Args:
        page: The page.
        link_class: The CSS class(es) for the link.
    """
    page_url = page.url

    link_class_html = f' class="{link_class}"' if link_class else ""

    return f'<a href="{page_url}" title="View the {page.title} page"{link_class_html}>{page.title}</a>'


def post_read_more_link(
    post: Post,
    link_class: str = "",
    read_more_text: str = "",
) -> str:
    """Return the read more link for a post.

    Args:
        post: The post.
        link_class: The CSS class(es) for the link.
        read_more_text: The text for the read more link.

    Returns:
        str: The read more link.
    """
    read_more_text = read_more_text if read_more_text else djpress_settings.POST_READ_MORE_TEXT
    link_class_html = f' class="{link_class}"' if link_class else ""

    return f'<p><a href="{post.url}"{link_class_html}>{read_more_text}</a></p>'


def get_blog_pages_list(
    pages: list[dict[Post, list]],
    li_class: str = "",
    a_class: str = "",
    ul_child_class: str = "",
) -> str:
    """Return the HTML for the blog pages list.

    This expects to be passed the output of the get_page_tree method.

    Args:
        pages: The pages from the get_page_tree method.
        li_class: The CSS class(es) for the list item.
        a_class: The CSS class(es) for the link.
        ul_child_class: The CSS class(es) for the child ul

    Returns:
        str: The HTML for the blog pages list.
    """
    class_li = f' class="{li_class}"' if li_class else ""
    class_ul = f' class="{ul_child_class}"' if ul_child_class else ""

    output = ""

    for page_data in pages:
        page = page_data["page"]
        children = page_data["children"]

        output += f"<li{class_li}>{get_page_link(page, link_class=a_class)}"

        if children:
            output += f"<ul{class_ul}>"
            output += get_blog_pages_list(children, li_class=li_class, a_class=a_class, ul_child_class=ul_child_class)
            output += "</ul>"

        output += "</li>"

    return output


class BlogPostWrapper(template.Node):
    """Wraps the blog post content.

    This is a template tag that wraps the blog post content in a configurable HTML tag with a CSS class.

    Args:
        nodelist: The content to wrap.
        tag: The HTML tag to wrap the content in.
        css_class: The CSS class(es) for the tag.

    Returns:
        str: The wrapped content.
    """

    def __init__(self, nodelist: template.NodeList, tag: str = "", css_class: str = "") -> None:
        """Initialize the BlogPostWrapper."""
        self.nodelist = nodelist
        self.tag = "article" if tag == "" else tag
        self.css_class = css_class

    def render(self, context: template.Context) -> str:
        """Render the blog post content."""
        content = self.nodelist.render(context)

        # Just return the content if the tag isn't allowed
        if self.tag not in ["div", "span", "section", "article"]:
            return mark_safe(content)

        return mark_safe(f"<{self.tag}{self.css_class}>{content}</{self.tag}>")


def parse_post_wrapper_params(params: list) -> tuple[str, str]:
    """Parse the parameters for the template tag.

    Args:
        params: The parameters for the template tag.

    Returns:
        tuple: The tag and CSS class.
    """
    tag = ""
    css_class = ""

    for num, param in enumerate(params):
        # Support keyword arguments
        if "=" in param:
            name, value = param.split("=", 1)
            value = value.strip("\"'")
            if name == "tag":
                tag = value
            elif name == "class":
                css_class = value
            else:
                # Ignore any other keyword arguments
                pass
        else:
            # Support positional arguments too
            value = param.strip("\"'")
            if num == 0:
                tag = value
            if num == 1:
                css_class = value

    return tag, css_class



import pytest

from djpress.sitemaps import DateBasedSitemap, PostSitemap, PageSitemap, CategorySitemap
from djpress.url_utils import get_archives_url, get_category_url
from djpress.models import Post


@pytest.mark.django_db
def test_post_sitemap(test_post1, test_post2, test_post3):
    """Test the PostSitemap class."""

    expected_items = [test_post1, test_post2, test_post3]

    sitemap = PostSitemap()

    assert sitemap.changefreq == "monthly"
    assert sitemap.protocol == "https"
    assert len(sitemap.items()) == len(expected_items)
    assert sitemap.lastmod(test_post1) == test_post1.modified_date
    assert sitemap.location(test_post1) == test_post1.url
    assert sitemap.lastmod(test_post2) == test_post2.modified_date
    assert sitemap.location(test_post2) == test_post2.url
    assert sitemap.lastmod(test_post3) == test_post3.modified_date
    assert sitemap.location(test_post3) == test_post3.url


@pytest.mark.django_db
def test_page_sitemap(test_page1, test_page2, test_page3):
    """Test the PageSitemap class."""

    expected_items = [test_page1, test_page2, test_page3]

    sitemap = PageSitemap()

    assert sitemap.changefreq == "monthly"
    assert sitemap.protocol == "https"
    assert len(sitemap.items()) == len(expected_items)
    assert sitemap.lastmod(test_page1) == test_page1.modified_date
    assert sitemap.location(test_page1) == test_page1.url
    assert sitemap.lastmod(test_page2) == test_page2.modified_date
    assert sitemap.location(test_page2) == test_page2.url
    assert sitemap.lastmod(test_page3) == test_page3.modified_date
    assert sitemap.location(test_page3) == test_page3.url


@pytest.mark.django_db
def test_category_sitemap(category1, category2, test_post1, test_post2):
    """Test the CategorySitemap class."""

    expected_items = [category1, category2]

    sitemap = CategorySitemap()

    assert sitemap.changefreq == "daily"
    assert sitemap.protocol == "https"
    assert len(sitemap.items()) == len(expected_items)
    assert sitemap.lastmod(category1) == test_post1.modified_date
    assert sitemap.lastmod(category2) == test_post2.modified_date
    assert sitemap.location(category1) == get_category_url(category1)


@pytest.mark.django_db
def test_date_based_sitemap(test_post1, test_post2, test_post3):
    """Test the DateBasedSitemap class."""

    expected_items = [
        {"year": test_post3.date.year, "latest_modified": test_post3.modified_date},
        {"year": test_post3.date.year, "month": test_post3.date.month, "latest_modified": test_post3.modified_date},
        {
            "year": test_post3.date.year,
            "month": test_post3.date.month,
            "day": test_post3.date.day,
            "latest_modified": test_post3.modified_date,
        },
    ]

    sitemap = DateBasedSitemap()

    assert sitemap.changefreq == "daily"
    assert sitemap.protocol == "https"
    assert sitemap.items() == expected_items
    assert sitemap.lastmod(expected_items[0]) == test_post3.modified_date
    assert sitemap.location(expected_items[0]) == get_archives_url(test_post3.date.year)
    assert sitemap.location(expected_items[1]) == get_archives_url(test_post3.date.year, test_post3.date.month)
    assert sitemap.location(expected_items[2]) == get_archives_url(
        test_post3.date.year, test_post3.date.month, test_post3.date.day
    )


@pytest.mark.django_db
def test_date_based_sitemap_archives_disabled(settings, test_post1, test_post2, test_post3):
    # Check that we have three published posts
    assert Post.post_objects.count() == 3

    # Disable the archives
    settings.DJPRESS_SETTINGS["ARCHIVE_ENABLED"] = False

    # Create a new sitemap
    sitemap = DateBasedSitemap()

    # Check that the items are empty
    assert sitemap.items() == []



"""Configuration file for the Sphinx documentation builder."""

from pathlib import Path

import tomllib

# Define the path to pyproject.toml
pyproject_path = Path(__file__).parent.parent / "pyproject.toml"

# Read the version from pyproject.toml
with pyproject_path.open("rb") as f:
    pyproject_data = tomllib.load(f)
    djpress_version = pyproject_data["project"]["version"]

# General information about the project.
project = "DJ Press"
copyright = "2024, Stuart Maxwell"  # noqa: A001
author = "Stuart Maxwell"
release = djpress_version

# -- General configuration ---------------------------------------------------

extensions = [
    "myst_parser",
]
myst_enable_extensions = ["colon_fence"]
myst_heading_anchors = 3

templates_path = ["_templates"]
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
]

language = "en"

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "nord"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_theme_options = {}
html_title = project
html_static_path = []



import pytest
from django.contrib.auth.models import User
from django.urls import reverse

from djpress.conf import settings as djpress_settings
from djpress.models import Post
from djpress.url_utils import get_rss_url


@pytest.mark.django_db
def test_latest_posts_feed(client, user):
    Post.post_objects.create(title="Post 1", content="Content of post 1.", author=user, status="published")
    Post.post_objects.create(title="Post 2", content="Content of post 2.", author=user, status="published")

    url = get_rss_url()
    print(f"URL: {url}")
    response = client.get(url)

    assert response.status_code == 200
    assert response["Content-Type"] == "application/rss+xml; charset=utf-8"

    feed = response.content.decode("utf-8")
    assert f"<title>{djpress_settings.BLOG_TITLE}</title>" in feed
    assert f"<link>http://testserver/{djpress_settings.RSS_PATH}/</link>" in feed
    assert f"<description>{djpress_settings.BLOG_DESCRIPTION}</description>" in feed
    assert "<item>" in feed
    assert "<title>Post 1</title>" in feed
    assert "<description>&lt;p&gt;Content of post 1.&lt;/p&gt;</description>" in feed
    assert "<title>Post 2</title>" in feed
    assert "<description>&lt;p&gt;Content of post 2.&lt;/p&gt;</description>" in feed


@pytest.mark.django_db
def test_truncated_posts_feed(client, user):
    # Confirm the truncate tag is set according to settings_testing.py
    truncate_tag = "<!--test-more-->"
    assert djpress_settings.TRUNCATE_TAG == truncate_tag
    post_prefix = "test-posts"
    assert djpress_settings.POST_PREFIX == post_prefix

    Post.post_objects.create(
        title="Post 1",
        content=f"Content of post 1.{truncate_tag}Truncated content",
        author=user,
        status="published",
    )

    url = get_rss_url()
    response = client.get(url)

    assert response.status_code == 200
    assert response["Content-Type"] == "application/rss+xml; charset=utf-8"

    feed = response.content.decode("utf-8")
    assert f"<title>{djpress_settings.BLOG_TITLE}</title>" in feed
    assert f"<link>http://testserver/{djpress_settings.RSS_PATH}/</link>" in feed
    assert f"<description>{djpress_settings.BLOG_DESCRIPTION}</description>" in feed
    assert "<item>" in feed
    assert "<title>Post 1</title>" in feed
    assert "Truncated content" not in feed
    assert f'&lt;a href="/{post_prefix}/post-1/"&gt;Read more&lt;/a&gt;&lt;/p&gt;' in feed


@pytest.mark.django_db
def test_pages_not_in_feed(client, test_post1, test_post2, test_post3, test_page1, user):
    url = get_rss_url()
    response = client.get(url)

    assert response.status_code == 200
    assert response["Content-Type"] == "application/rss+xml; charset=utf-8"

    feed = response.content.decode("utf-8")
    assert "<item>" in feed
    assert test_post1.title in feed
    assert test_post2.title in feed
    assert test_post3.title in feed
    assert test_page1.title not in feed

    Post.page_objects.create(
        title="New Test Page",
        content="Content of page 1.",
        author=user,
        status="published",
        post_type="page",
    )

    response = client.get(url)

    assert response.status_code == 200
    assert response["Content-Type"] == "application/rss+xml; charset=utf-8"

    feed = response.content.decode("utf-8")
    assert "<item>" in feed
    assert "New Test Page" not in feed



"""Custom checks for DJPress."""

from django.core.checks import Warning


def check_plugin_hooks(app_configs, **kwargs) -> list[Warning]:  # noqa: ANN001, ANN003, ARG001
    """Check for unknown plugin hooks."""
    from djpress.plugins import Hooks, registry

    # Ensure plugins are loaded
    if not registry._loaded:  # noqa: SLF001
        registry.load_plugins()

    warnings = []

    for hook_name in registry.hooks:
        if not isinstance(hook_name, Hooks):
            warning = Warning(
                f"Plugin registering unknown hook '{hook_name}'.",
                hint=("This might indicate use of a deprecated hook or a hook from a newer version of DJPress."),
                obj=hook_name,
                id="djpress.W001",
            )
            warnings.append(warning)

    return warnings



"""URL patterns for the djpress application."""

import re

from django.conf import settings as django_settings
from django.contrib.auth.models import User

from djpress.conf import settings as djpress_settings
from djpress.models.post import Category, Post


def regex_post() -> str:
    """Convert the post prefix to a regex pattern.

    This will match the following URL parts:
    - The year in the format of 4 digits.
    - The month in the format of 2 digits.
    - The day in the format of 2 digits.
    - The slug in the format of word characters or hyphens.

    Args:
        prefix (str): The post prefix that is configured in the settings.

    Returns:
        str: The regex pattern.
    """
    prefix = djpress_settings.POST_PREFIX

    regex_parts = []

    # Regexes are complicated - this is what the following does:
    # - `(...)`: The parentheses create a capturing group. This means that the splits will occur around these matches,
    #   but the matches themselves will be included in the resulting list.
    # - `\{\{`: This matches two opening curly braces {{. The backslashes are necessary because curly braces have
    #   special meaning in regex, so we need to escape them to match literal curly braces.
    # - `.*?`: This is matches the characters inside the curly brackets.
    #   - `.`: Matches any single character.
    #   - `*`: Means "zero or more" of the preceding pattern.
    #   - `?`: Makes the `*` non-greedy, meaning it will match as few characters as possible.
    # - `\}\}`: This matches two closing curly braces }}, again escaped with backslashes.
    parts = re.split(r"(\{\{.*?\}\})", prefix)

    for part in parts:
        # Remove spaces from the part so that either {{ year }} or {{year}} will work
        if part.replace(" ", "") == "{{year}}":
            regex_parts.append(r"(?P<year>\d{4})")
        elif part.replace(" ", "") == "{{month}}":
            regex_parts.append(r"(?P<month>\d{2})")
        elif part.replace(" ", "") == "{{day}}":
            regex_parts.append(r"(?P<day>\d{2})")
        else:
            # Escape the part, but replace escaped spaces with regular spaces
            escaped_part = re.escape(part).replace("\\ ", " ")
            regex_parts.append(escaped_part)

    regex = "".join(regex_parts)

    # If the regex is blank we return just the slug re, otherwise we append a slash and the slug re
    # Regex explanation:
    # - (?P<slug>...): This is a named capture group.
    # - [\w-]: This matches any word character (alphanumeric or underscore) or a hyphen.
    # - +: This means "one or more" of the preceding pattern.
    if not regex:
        return r"(?P<slug>[\w-]+)"

    return rf"{regex}/(?P<slug>[\w-]+)"


def get_post_url(post: Post) -> str:
    """Return the URL for the post."""
    prefix = djpress_settings.POST_PREFIX

    # Remove spaces from the prefix so that either {{ year }} or {{year}} will work
    prefix = prefix.replace(" ", "")

    # Replace the placeholders in the prefix with the actual values
    if "{{year}}" in prefix:
        prefix = prefix.replace("{{year}}", post.date.strftime("%Y"))
    if "{{month}}" in prefix:
        prefix = prefix.replace("{{month}}", post.date.strftime("%m"))
    if "{{day}}" in prefix:
        prefix = prefix.replace("{{day}}", post.date.strftime("%d"))

    url = f"/{post.slug}" if prefix == "" else f"/{prefix}/{post.slug}"

    if django_settings.APPEND_SLASH:
        return f"{url}/"

    return url


def regex_archives() -> str:
    """Generate the regex path for the archives view.

    The following regex is used to match the archives path. It is used to match
    the following patterns:
    - 2024
    - 2024/01
    - 2024/01/01

    Returns:
        str: The regex pattern.
    """
    # Regex explanation:
    # - (?P<year>\d{4}): Required - this matches the year in the format of 4 digits.
    # - (/(?P<month>\d{2}))?: Optional - this matches the month in the format of 2 digits.
    # - (/(?P<day>\d{2}))?: Optional - this matches the day in the format of 2 digits.
    regex = r"(?P<year>\d{4})(/(?P<month>\d{2}))?(/(?P<day>\d{2}))?"

    if djpress_settings.ARCHIVE_PREFIX:
        regex = rf"{re.escape(djpress_settings.ARCHIVE_PREFIX)}/{regex}"

    return regex


def regex_page() -> str:
    """Generate the regex path for pages.

    The following regex is used to match the path. It is used to match the
    any path that contains letters, numbers, underscores, hyphens, and slashes.
    """
    # Regex explanation:
    # - (?P<path>([\w-]+/)*[\w-]+): This matches the path.
    #   - (?P<path>...): This is a named capture group.
    #   - ([\w-]+/)*: This matches any word character (alphanumeric or underscore) or a hyphen, followed by a slash.
    #     - [\w-]: This matches any word character (alphanumeric or underscore) or a hyphen.
    #     - +: This means "one or more" of the preceding pattern.
    return r"(?P<path>([\w-]+/)*[\w-]+)"


def regex_category() -> str:
    """Generate the regex path for the category view."""
    # Regex explanation:
    # - (?P<slug>[\w-]+): This is a named capture group that matches any word character (alphanumeric or underscore)
    #   or a hyphen.
    regex = r"(?P<slug>[\w-]+)"

    if djpress_settings.CATEGORY_PREFIX:
        regex = rf"{re.escape(djpress_settings.CATEGORY_PREFIX)}/{regex}"

    return regex


def regex_author() -> str:
    """Generate the regex path for the author view."""
    # Regex explanation:
    # - (?P<author>[\w-]+): This is a named capture group that matches any word character (alphanumeric or underscore)
    #   or a hyphen.
    regex = r"(?P<author>[\w-]+)"

    if djpress_settings.AUTHOR_PREFIX:
        regex = rf"{re.escape(djpress_settings.AUTHOR_PREFIX)}/{regex}"

    return regex


def get_path_regex(path_match: str) -> str:
    """Return the regex for the requested match."""
    if path_match == "post":
        regex = regex_post()
    if path_match == "archives":
        regex = regex_archives()
    if path_match == "page":
        regex = regex_page()
    if path_match == "category":
        regex = regex_category()
    if path_match == "author":
        regex = regex_author()

    if django_settings.APPEND_SLASH:
        return f"^{regex}/$"

    return f"^{regex}$"


def get_author_url(user: User) -> str:
    """Return the URL for the author's page."""
    url = (
        f"/{djpress_settings.AUTHOR_PREFIX}/{user.username}"
        if djpress_settings.AUTHOR_PREFIX
        else f"/author/{user.username}"
    )

    if django_settings.APPEND_SLASH:
        return f"{url}/"

    return url


def get_category_url(category: "Category") -> str:
    """Return the URL for the category."""
    url = f"/{category.permalink}"

    if django_settings.APPEND_SLASH:
        return f"{url}/"

    return url


def get_archives_url(year: int, month: int | None = None, day: int | None = None) -> str:
    """Return the URL for the archives page."""
    url = f"/{djpress_settings.ARCHIVE_PREFIX}/{year}" if djpress_settings.ARCHIVE_PREFIX else f"/{year}"

    if month:
        url += f"/{month:02d}"
    if day:
        url += f"/{day:02d}"

    if django_settings.APPEND_SLASH:
        return f"{url}/"

    return url


def get_page_url(page: Post) -> str:
    """Return the URL for the page."""
    url = f"/{page.permalink}"

    if django_settings.APPEND_SLASH:
        return f"{url}/"

    return url


def get_rss_url() -> str:
    """Return the URL for the RSS feed.

    If the RSS path is not set, the default is "/feed". This will raise an error if the path is not set.

    Returns:
        str: The URL for the RSS feed.
    """
    url = f"/{djpress_settings.RSS_PATH}"

    if django_settings.APPEND_SLASH:
        return f"{url}/"

    return url



import pytest
from copy import deepcopy

from django.utils import timezone
from django.contrib.auth.models import User

from djpress.url_converters import SlugPathConverter
from djpress.models import Category, Post
from djpress.plugins import DJPressPlugin, registry

from example.config import settings_testing

# Take a static snapshot of DJPRESS_SETTINGS from settings_test.py
CLEAN_DJPRESS_SETTINGS = deepcopy(settings_testing.DJPRESS_SETTINGS)


@pytest.fixture(autouse=True)
def reset_djpress_settings(settings):
    """Reset DJPress settings for each test based on a clean, static state."""

    # Reset to the known clean state
    settings.DJPRESS_SETTINGS.clear()
    settings.DJPRESS_SETTINGS.update(CLEAN_DJPRESS_SETTINGS)

    yield  # Run the test

    # Ensure everything is cleared again after the test (extra cleanup)
    settings.DJPRESS_SETTINGS.clear()
    settings.DJPRESS_SETTINGS.update(CLEAN_DJPRESS_SETTINGS)


@pytest.fixture
def converter():
    return SlugPathConverter()


@pytest.fixture
def mock_timezone_now(monkeypatch):
    current_time = timezone.now()
    monkeypatch.setattr(timezone, "now", lambda: current_time)
    return current_time


@pytest.fixture
def user():
    return User.objects.create_user(
        username="testuser",
        password="testpass",
        first_name="Test",
        last_name="User",
    )


@pytest.fixture
def category1():
    return Category.objects.create(title="Test Category1", slug="test-category1")


@pytest.fixture
def category2():
    return Category.objects.create(title="Test Category2", slug="test-category2")


@pytest.fixture
def category3():
    category = Category.objects.create(title="Development", slug="dev")
    return category


@pytest.fixture
def test_post1(user, category1):
    post = Post.objects.create(
        title="Test Post1",
        slug="test-post1",
        content="This is test post 1.",
        author=user,
        status="published",
        post_type="post",
    )
    post.categories.set([category1])

    return post


@pytest.fixture
def test_post2(user, category2):
    post = Post.objects.create(
        title="Test Post2",
        slug="test-post2",
        content="This is test post 2.",
        author=user,
        status="published",
        post_type="post",
    )
    post.categories.set([category2])
    return post


@pytest.fixture
def test_post3(user, category1):
    post = Post.objects.create(
        title="Test Post3",
        slug="test-post3",
        content="This is test post 3.",
        author=user,
        status="published",
        post_type="post",
    )

    return post


@pytest.fixture
def test_long_post1(user, settings, category1):
    truncate_tag = settings.DJPRESS_SETTINGS["TRUNCATE_TAG"]
    post = Post.post_objects.create(
        title="Test Long Post1",
        slug="test-long-post1",
        content=f"This is the truncated content.\n\n{truncate_tag}\n\nThis is the rest of the post.",
        author=user,
        status="published",
        post_type="post",
    )
    post.categories.set([category1])
    return post


@pytest.fixture
def test_page1(user):
    return Post.objects.create(
        title="Test Page1",
        slug="test-page1",
        content="This is test page 1.",
        author=user,
        status="published",
        post_type="page",
    )


@pytest.fixture
def test_page2(user):
    return Post.objects.create(
        title="Test Page2",
        slug="test-page2",
        content="This is test page 2.",
        author=user,
        status="published",
        post_type="page",
    )


@pytest.fixture
def test_page3(user):
    return Post.objects.create(
        title="Test Page3",
        slug="test-page3",
        content="This is test page 3.",
        author=user,
        status="published",
        post_type="page",
    )


@pytest.fixture
def test_page4(user):
    return Post.objects.create(
        title="Test Page4",
        slug="test-page4",
        content="This is test page 4.",
        author=user,
        status="published",
        post_type="page",
    )


@pytest.fixture
def test_page5(user):
    return Post.objects.create(
        title="Test Page5",
        slug="test-page5",
        content="This is test page 5.",
        author=user,
        status="published",
        post_type="page",
    )


@pytest.fixture
def clean_registry():
    """Reset the plugin registry before and after each test."""
    # Reset before test
    registry.hooks = {}
    registry.plugins = []
    registry._loaded = False

    yield

    # Reset after test
    registry.hooks = {}
    registry.plugins = []
    registry._loaded = False


@pytest.fixture
def bad_plugin_registry(clean_registry):
    """Create a registry with a plugin that uses an unknown hook."""

    class BadPlugin(DJPressPlugin):
        name = "bad_plugin"

        def setup(self, registry):
            registry.register_hook("foobar", lambda x: x)

    # Manually create and setup the plugin
    plugin = BadPlugin()
    plugin.setup(registry)

    return registry



"""Default Markdown renderer for Djpress."""

import markdown

from djpress.conf import settings as djpress_settings


def default_renderer(markdown_text: str) -> str:
    """Return the Markdown text as HTML."""
    md = markdown.Markdown(
        extensions=djpress_settings.MARKDOWN_EXTENSIONS,
        extension_configs=djpress_settings.MARKDOWN_EXTENSION_CONFIGS,
        output_format="html",
    )
    html = md.convert(markdown_text)
    md.reset()

    return html



import pytest
from djpress.plugins import PluginRegistry, DJPressPlugin, Hooks
from djpress.plugins import registry

from djpress.conf import settings as djpress_settings


def test_registry_initialization():
    """Test that registry initializes with empty plugins and hooks."""
    registry = PluginRegistry()
    assert registry.plugins == []
    assert registry.hooks == {}
    assert registry._loaded is False


def test_register_hook():
    """Test registering a hook callback."""
    registry = PluginRegistry()

    def test_callback(content: str) -> str:
        return content

    registry.register_hook(Hooks.PRE_RENDER_CONTENT, test_callback)
    assert test_callback in registry.hooks[Hooks.PRE_RENDER_CONTENT]


def test_hook_enum_values():
    """Test that Hook enum has the expected values."""
    assert Hooks.PRE_RENDER_CONTENT.value == "pre_render_content"
    assert Hooks.POST_RENDER_CONTENT.value == "post_render_content"


# Tests for plugin registration (flexible)
def test_register_hook_with_enum(clean_registry):
    """Test registering hook using Enum."""

    def test_callback(content: str) -> str:
        return content

    registry.register_hook(Hooks.PRE_RENDER_CONTENT, test_callback)
    assert Hooks.PRE_RENDER_CONTENT in registry.hooks
    assert test_callback in registry.hooks[Hooks.PRE_RENDER_CONTENT]


def test_register_hook_with_valid_string(clean_registry):
    """Test registering hook using valid string name."""

    def test_callback(content: str) -> str:
        return content

    registry.register_hook("pre_render_content", test_callback)
    assert Hooks.PRE_RENDER_CONTENT in registry.hooks
    assert test_callback in registry.hooks[Hooks.PRE_RENDER_CONTENT]


def test_register_unknown_hook_name(clean_registry):
    """Test registering unknown hook name - should accept but trigger warning."""

    def test_callback(content: str) -> str:
        return content

    registry.register_hook("unknown_hook", test_callback)
    assert "unknown_hook" in registry.hooks
    assert test_callback in registry.hooks["unknown_hook"]


# Tests for hook execution (strict)
def test_run_hook_with_enum(clean_registry):
    """Test running hook with Enum succeeds."""

    def test_callback(content: str) -> str:
        return content + " modified"

    registry.register_hook(Hooks.PRE_RENDER_CONTENT, test_callback)
    result = registry.run_hook(Hooks.PRE_RENDER_CONTENT, "test")
    assert result == "test modified"


def test_run_hook_with_string_fails(clean_registry):
    """Test running hook with string fails."""
    with pytest.raises(TypeError, match="hook_name must be a Hooks enum member"):
        registry.run_hook("pre_render_content", "test")


def test_run_hook_type_error_message(clean_registry):
    """Test specific error message when running hook with wrong type."""
    with pytest.raises(TypeError) as exc_info:
        registry.run_hook("pre_render_content", "test")
    assert "hook_name must be a Hooks enum member" in str(exc_info.value)
    assert "got <class 'str'>" in str(exc_info.value)


# For lines 48-50 (contextlib.suppress for ValueError)
def test_register_hook_with_invalid_string(clean_registry):
    """Test registering hook with invalid string - should store as string."""

    def test_callback(content: str) -> str:
        return content

    registry.register_hook("not_a_real_hook", test_callback)
    # Should store with string key since conversion to Enum failed
    assert "not_a_real_hook" in registry.hooks
    assert test_callback in registry.hooks["not_a_real_hook"]


# For line 90 (value is None in hooks dict)
def test_run_hook_nonexistent_hook(clean_registry):
    """Test running hook that hasn't been registered."""
    result = registry.run_hook(Hooks.PRE_RENDER_CONTENT, "test")
    assert result == "test"  # Should return original value unchanged


# For lines 96-98 (plugin loading already done)
def test_load_plugins_already_loaded(clean_registry):
    """Test that load_plugins doesn't reload if already loaded."""
    registry._loaded = True
    registry.load_plugins()
    assert len(registry.plugins) == 0  # Should not have loaded any plugins


# For lines 115-131 (_import_plugin_class)
def test_import_plugin_standard_location(clean_registry, tmp_path):
    """Test importing plugin from standard location."""
    # Create a temporary plugin package
    plugin_dir = tmp_path / "test_plugin"
    plugin_dir.mkdir()
    (plugin_dir / "__init__.py").write_text("")
    (plugin_dir / "plugin.py").write_text("""
from djpress.plugins import DJPressPlugin
class Plugin(DJPressPlugin):
    name = "test_plugin"
    """)

    # Add to Python path and try to import
    import sys

    sys.path.insert(0, str(tmp_path))

    try:
        plugin_class = registry._import_plugin_class("test_plugin")
        assert plugin_class.__name__ == "Plugin"
    finally:
        sys.path.pop(0)


def test_import_plugin_custom_location(clean_registry, tmp_path):
    """Test importing plugin from custom location."""
    # Similar setup but with custom path
    # Test both successful and failing imports


# For lines 147-165 and 175-178 (_instantiate_plugin)
def test_instantiate_plugin_with_config(clean_registry):
    """Test instantiating plugin with configuration."""

    class TestPlugin(DJPressPlugin):
        name = "test_plugin"

    plugin_settings = {"test_plugin": {"setting": "value"}}
    plugin = registry._instantiate_plugin(TestPlugin, "test_plugin", plugin_settings)
    assert isinstance(plugin, DJPressPlugin)
    assert plugin.config == {"setting": "value"}


def test_instantiate_plugin_fails(clean_registry):
    """Test handling of plugin instantiation failure."""

    class BrokenPlugin(DJPressPlugin):
        name = "broken_plugin"

        def __init__(self, config=None):
            raise RuntimeError("Plugin broken")

    from django.core.exceptions import ImproperlyConfigured

    with pytest.raises(ImproperlyConfigured) as exc_info:
        registry._instantiate_plugin(BrokenPlugin, "broken_plugin", {})
    assert "Error initializing plugin" in str(exc_info.value)


# For lines 48-50 (contextlib.suppress and string conversion)
def test_register_hook_conversion_from_string(clean_registry):
    """Test string to enum conversion in register_hook."""

    def callback(content: str) -> str:
        return content

    # This should try to convert the string to enum and succeed
    registry.register_hook("pre_render_content", callback)
    assert Hooks.PRE_RENDER_CONTENT in registry.hooks

    # This should try to convert and silently fail
    registry.register_hook("not_a_hook", callback)
    assert "not_a_hook" in registry.hooks


# For lines 96-98 (early return if already loaded)
def test_load_plugins_early_return(clean_registry):
    """Test that load_plugins returns early if already loaded."""
    registry._loaded = True
    original_plugins = registry.plugins.copy()
    registry.load_plugins()
    assert registry.plugins == original_plugins


# For lines 123-131 (import error handling)
def test_import_plugin_both_paths_fail(clean_registry):
    """Test when both import attempts fail."""
    from django.core.exceptions import ImproperlyConfigured

    with pytest.raises(ImproperlyConfigured) as exc_info:
        registry._import_plugin_class("nonexistent_plugin")

    assert "Could not load plugin 'nonexistent_plugin'" in str(exc_info.value)
    assert "Tried both custom path and standard plugin.py location" in str(exc_info.value)


# For lines 176-177 (plugin instantiation error)
def test_instantiate_plugin_setup_fails(clean_registry):
    """Test handling of plugin setup failure."""

    class PluginWithBadSetup(DJPressPlugin):
        name = "bad_setup"

        def setup(self, registry):
            raise RuntimeError("Setup failed")

    from django.core.exceptions import ImproperlyConfigured

    with pytest.raises(ImproperlyConfigured) as exc_info:
        registry._instantiate_plugin(PluginWithBadSetup, "bad_setup", {})

    assert "Error initializing plugin 'bad_setup'" in str(exc_info.value)


def test_register_hook_first_callback(clean_registry):
    """Test registering first callback for a hook."""

    def callback(content: str) -> str:
        return content

    # This should create new list and add callback
    registry.register_hook(Hooks.PRE_RENDER_CONTENT, callback)
    assert len(registry.hooks[Hooks.PRE_RENDER_CONTENT]) == 1
    assert registry.hooks[Hooks.PRE_RENDER_CONTENT][0] == callback


def test_load_plugins_successful(clean_registry, monkeypatch):
    """Test successful plugin loading process."""

    class TestPlugin(DJPressPlugin):
        name = "test"

        def setup(self, registry):
            pass

    # Mock the import and instantiate methods to return known values
    def mock_import(self, path):
        return TestPlugin

    def mock_instantiate(self, plugin_class, path, settings):
        return TestPlugin()

    monkeypatch.setattr(PluginRegistry, "_import_plugin_class", mock_import)
    monkeypatch.setattr(PluginRegistry, "_instantiate_plugin", mock_instantiate)

    # Set up test settings
    monkeypatch.setattr(djpress_settings, "PLUGINS", ["test_plugin"])
    monkeypatch.setattr(djpress_settings, "PLUGIN_SETTINGS", {})

    # Load plugins
    registry.load_plugins()

    # Verify plugin was added
    assert len(registry.plugins) == 1
    assert isinstance(registry.plugins[0], TestPlugin)


def test_plugin_missing_name():
    """Test plugin initialization with missing name."""

    class NoNamePlugin(DJPressPlugin):
        pass  # No name defined

    with pytest.raises(ValueError) as exc_info:
        NoNamePlugin()
    assert str(exc_info.value) == "Plugin must define a name"


def test_register_multiple_callbacks(clean_registry):
    """Test registering multiple callbacks for the same hook."""

    def callback1(content: str) -> str:
        return content + "1"

    def callback2(content: str) -> str:
        return content + "2"

    # Register first callback - should create new list
    registry.register_hook(Hooks.PRE_RENDER_CONTENT, callback1)
    assert len(registry.hooks[Hooks.PRE_RENDER_CONTENT]) == 1
    assert registry.hooks[Hooks.PRE_RENDER_CONTENT][0] == callback1

    # Register second callback - should append to existing list
    registry.register_hook(Hooks.PRE_RENDER_CONTENT, callback2)
    assert len(registry.hooks[Hooks.PRE_RENDER_CONTENT]) == 2
    assert registry.hooks[Hooks.PRE_RENDER_CONTENT][0] == callback1
    assert registry.hooks[Hooks.PRE_RENDER_CONTENT][1] == callback2

    # Verify both callbacks are executed in order
    result = registry.run_hook(Hooks.PRE_RENDER_CONTENT, "test")
    assert result == "test12"


def test_content_modification_hook(clean_registry):
    """Test hooks that modify and return content."""

    def test_callback(content: str) -> str:
        return content + " modified"

    registry.register_hook(Hooks.PRE_RENDER_CONTENT, test_callback)
    result = registry.run_hook(Hooks.PRE_RENDER_CONTENT, "test")
    assert result == "test modified"


@pytest.mark.django_db
def test_plugin_storage_interface():
    """Test plugin storage interface methods."""

    class TestPlugin(DJPressPlugin):
        name = "test_plugin"

    plugin = TestPlugin()

    # Test get_data with no storage
    assert plugin.get_data() == {}

    # Test save_data and get_data
    plugin.save_data({"key": "value"})
    assert plugin.get_data() == {"key": "value"}

    # Test save_data with update
    plugin.save_data({"new_key": "new_value"})
    assert plugin.get_data() == {"new_key": "new_value"}

    # Test save additional data
    data = plugin.get_data()
    data["extra"] = "data"
    plugin.save_data(data)
    assert plugin.get_data() == {"new_key": "new_value", "extra": "data"}



import re
import pytest


def test_regex_append_slash(converter):
    # Valid paths
    valid_paths = [
        "valid-path/",
        "valid_path/",
        "valid/path/",
        "valid123/",
        "valid-123_path/",
        "valid/path/123/",
    ]
    for path in valid_paths:
        assert re.match(converter.regex, path), f"Regex should match valid path: {path}"

    # Invalid paths
    invalid_paths = [
        "invalid path",  # space
        "invalid@path",  # special character
        "invalid#path",  # special character
        "invalid?path",  # special character
    ]
    for path in invalid_paths:
        assert not re.match(converter.regex, path), f"Regex should not match invalid path: {path}"


def test_regex_no_append_slash(settings, converter):
    settings.APPEND_SLASH = False
    # Valid paths
    valid_paths = [
        "valid-path",
    ]
    for path in valid_paths:
        assert re.match(converter.regex, path), f"Regex should match valid path: {path}"


def test_to_python(converter):
    assert converter.to_python("test-value") == "test-value"
    assert converter.to_python("123") == "123"
    assert converter.to_python("valid/path") == "valid/path"


def test_to_url(converter):
    assert converter.to_url("test-value") == "test-value"
    assert converter.to_url("123") == "123"
    assert converter.to_url("valid/path") == "valid/path"



"""Category model."""

from django.core.cache import cache
from django.db import IntegrityError, models, transaction
from django.db.models import Max
from django.utils import timezone
from django.utils.text import slugify

from djpress.conf import settings as djpress_settings

CATEGORY_CACHE_KEY = "categories"


class CategoryManager(models.Manager):
    """Category manager."""

    def get_categories(self: "CategoryManager") -> models.QuerySet:
        """Return the queryset for categories.

        If CACHE_CATEGORIES is set to True, we return the cached queryset.
        """
        if djpress_settings.CACHE_CATEGORIES:
            return self._get_cached_categories()

        return self.all()

    def _get_cached_categories(self: "CategoryManager") -> models.QuerySet:
        """Return the cached categories queryset."""
        queryset = cache.get(CATEGORY_CACHE_KEY)

        if queryset is None:
            queryset = self.all()
            cache.set(CATEGORY_CACHE_KEY, queryset, timeout=None)

        return queryset

    def get_category_by_slug(self: "CategoryManager", slug: str) -> "Category":
        """Return a single category by its slug."""
        # First, try to get the category from the cache
        categories = self.get_categories()
        category = next(
            (category for category in categories if category.slug == slug),
            None,
        )

        # If the category is not found in the cache, fetch it from the database
        if not category:
            try:
                category = self.get(slug=slug)
            except Category.DoesNotExist as exc:
                msg = "Category not found"
                # Raise a 404 error
                raise ValueError(msg) from exc

        return category

    def get_categories_with_published_posts(self) -> "Category":
        """Return a queryset of categories that have published posts.

        We can use the has_posts property to include only categories with published posts.
        """
        return Category.objects.filter(pk__in=[category.pk for category in self.get_queryset() if category.has_posts])


class Category(models.Model):
    """Category model."""

    title = models.CharField(max_length=100)
    slug = models.SlugField(unique=True, blank=True)
    description = models.TextField(blank=True)
    menu_order = models.IntegerField(default=0)

    # Custom Manager
    objects: "CategoryManager" = CategoryManager()

    class Meta:
        """Meta options for the category model."""

        verbose_name = "category"
        verbose_name_plural = "categories"

    def __str__(self: "Category") -> str:
        """Return the string representation of the category."""
        return self.title

    def save(self: "Category", *args, **kwargs) -> None:  # noqa: ANN002, ANN003
        """Override the save method to auto-generate the slug."""
        if not self.slug:
            self.slug = slugify(self.title)
            if not self.slug or self.slug.strip("-") == "":
                msg = "Invalid title. Unable to generate a valid slug."
                raise ValueError(msg)

        try:
            with transaction.atomic():
                super().save(*args, **kwargs)
        except IntegrityError as exc:
            msg = f"A category with the slug {self.slug} already exists."
            raise ValueError(msg) from exc

    @property
    def permalink(self: "Category") -> str:
        """Return the category's permalink."""
        if djpress_settings.CATEGORY_ENABLED and djpress_settings.CATEGORY_PREFIX:
            return f"{djpress_settings.CATEGORY_PREFIX}/{self.slug}"

        return f"{self.slug}"

    @property
    def url(self) -> str:
        """Return the category's URL."""
        from djpress.url_utils import get_category_url

        return get_category_url(self)

    @property
    def posts(self) -> models.QuerySet:
        """Return only published posts."""
        return self._posts.filter(
            status="published",
            date__lte=timezone.now(),
        )

    @property
    def has_posts(self: "Category") -> bool:
        """Return True if the category has published posts."""
        return self.posts.exists()

    @property
    def last_modified(self: "Category") -> None | timezone.datetime:
        """Return the most recent last modified date of posts in the category.

        This property is used in the sitemap to determine the last modified date of the category.

        If the category has no published posts, we return None.

        Returns:
            None | timezone.datetime: The most recent last modified date of posts in the category.
        """
        return self.posts.aggregate(latest=Max("modified_date"))["latest"]



"""An example DJ Press plugin."""

from djpress.plugins import DJPressPlugin, PluginRegistry


class Plugin(DJPressPlugin):
    """An example DJ Press plugin."""

    name = "djpress_example_plugin"

    def setup(self, registry: PluginRegistry) -> None:
        """Set up the plugin.

        Args:
            registry (Hooks): The plugin registry.
        """
        registry.register_hook("pre_render_content", self.add_greeting)
        registry.register_hook("post_render_content", self.add_goodbye)

    def add_greeting(self, content: str) -> str:
        """Add a greeting to the content.

        This is a pre-render hook, so the content is still in Markdown format.

        Args:
            content (str): The content to modify.

        Returns:
            str: The modified content.
        """
        return f"{self.config.get("pre_text")} This was added by `djpress_example_plugin`!\n\n---\n\n{content}"

    def add_goodbye(self, content: str) -> str:
        """Add a goodbye message to the content.

        This is a post-render hook, so the content has already been rendered from Markdown to HTML.

        Args:
            content (str): The content to modify.

        Returns:
            str: The modified content.
        """
        return (
            f"{content}<hr><p>{self.config.get("pre_text")} This was added by <code>djpress_example_plugin</code>!</p>"
        )



"""Signals for djpress app."""

from django.core.cache import cache
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from djpress.models.category import (
    CATEGORY_CACHE_KEY,
    Category,
)
from djpress.models.post import (
    PUBLISHED_POSTS_CACHE_KEY,
    Post,
)


@receiver(post_save, sender=Category)
@receiver(post_delete, sender=Category)
def invalidate_category_cache(**_) -> None:  # noqa: ANN003
    """Invalidate the category cache.

    We invalidate the cache when a category is saved or deleted.
    """
    cache.delete(CATEGORY_CACHE_KEY)


@receiver(post_save, sender=Post)
@receiver(post_delete, sender=Post)
def invalidate_published_content_cache(**_) -> None:  # noqa: ANN003
    """Invalidate the published posts cache.

    We invalidate the cache when a post is saved or deleted.
    """
    cache.delete(PUBLISHED_POSTS_CACHE_KEY)



import pytest

from djpress.models import PluginStorage


@pytest.mark.django_db
def test_plugin_storage_model():
    test_plugin = PluginStorage.objects.create(plugin_name="test_plugin")

    assert str(test_plugin) == "test_plugin"


@pytest.mark.django_db
def test_plugin_storage_get_data():
    """Test getting plugin data."""
    # Test empty case (no storage exists)
    data = PluginStorage.objects.get_data("test_plugin")
    assert data == {}

    # Test with stored data
    PluginStorage.objects.save_data("test_plugin", {"key": "value"})
    data = PluginStorage.objects.get_data("test_plugin")
    assert data == {"key": "value"}


@pytest.mark.django_db
def test_plugin_storage_save_data():
    """Test saving plugin data."""
    # Test creating new storage
    PluginStorage.objects.save_data("test_plugin", {"key": "value"})
    storage = PluginStorage.objects.get(plugin_name="test_plugin")
    assert storage.plugin_data == {"key": "value"}

    # Test updating existing storage
    PluginStorage.objects.save_data("test_plugin", {"new_key": "new_value"})
    storage.refresh_from_db()
    assert storage.plugin_data == {"new_key": "new_value"}



import pytest
from django.contrib.auth.models import User
from django.core.paginator import Paginator
from django.template import Context, Template
from django.urls import reverse

from djpress.models import Category, Post
from djpress.templatetags import djpress_tags
from djpress.templatetags.helpers import (
    post_read_more_link,
    categories_html,
    get_page_link,
)
from djpress.utils import get_author_display_name
from djpress.exceptions import PageNotFoundError


@pytest.mark.django_db
def test_get_posts(test_post1, test_long_post1, test_post2, test_post3):
    posts = Post.objects.all().order_by("-date")

    assert list(djpress_tags.get_posts()) == list(posts)


@pytest.mark.django_db
def test_get_pages(test_page1, test_page2, test_page3):
    assert list(djpress_tags.get_pages()) == [test_page1, test_page2, test_page3]

    test_page1.status = "draft"
    test_page1.save()
    assert list(djpress_tags.get_pages()) == [test_page2, test_page3]

    test_page2.parent = test_page1
    test_page2.save()
    assert list(djpress_tags.get_pages()) == [test_page3]


@pytest.mark.django_db
def test_have_posts_single_post(test_post1):
    """Return a list of posts in the context."""
    context = Context({"post": test_post1})

    assert djpress_tags.have_posts(context) == [test_post1]


@pytest.mark.django_db
def test_have_posts_no_posts():
    """Return an empty list if there are no posts in the context."""
    context = Context({"foo": "bar"})

    assert djpress_tags.have_posts(context) == []


@pytest.mark.django_db
def test_have_posts_multiple_posts(test_post1, test_long_post1):
    """Return a list of posts in the context."""
    context = Context({"posts": [test_post1, test_long_post1]})

    assert djpress_tags.have_posts(context) == [test_post1, test_long_post1]


def test_blog_title(settings):
    """Test the blog_title template tag.

    This can be changed on the fly.
    """
    assert settings.DJPRESS_SETTINGS["BLOG_TITLE"] == "My Test DJ Press Blog"
    assert djpress_tags.blog_title() == settings.DJPRESS_SETTINGS["BLOG_TITLE"]

    # Change the title
    settings.DJPRESS_SETTINGS["BLOG_TITLE"] = "My New Blog Title"
    assert settings.DJPRESS_SETTINGS["BLOG_TITLE"] == "My New Blog Title"
    assert djpress_tags.blog_title() == settings.DJPRESS_SETTINGS["BLOG_TITLE"]


@pytest.mark.django_db
def test_get_categories(category1, category2, category3):
    categories = Category.objects.all().order_by("menu_order").order_by("title")
    djpress_categories = djpress_tags.get_categories()

    assert category1 in categories
    assert category2 in categories
    assert category3 in categories
    assert category1 in djpress_categories
    assert category2 in djpress_categories
    assert category3 in djpress_categories

    assert list(djpress_categories) == list(categories)  # type: ignore


@pytest.mark.django_db
def test_get_post_title_single_post(test_post1):
    context = Context({"post": test_post1})
    assert djpress_tags.get_post_title(context) == test_post1.title


def test_get_post_title_no_post_context():
    context = Context({"foo": "bar"})
    assert djpress_tags.get_post_title(context) == ""
    assert type(djpress_tags.get_post_title(context)) == str


@pytest.mark.django_db
def test_post_title_posts(settings, test_post1):
    """Test the post_title template tag.

    This uses the `post.permalink` property to generate the link."""
    # Context should have both a posts and a post to simulate the for post in posts loop
    context = Context({"posts": [test_post1], "post": test_post1})

    # Confirm settings in settings_testing.py
    assert settings.DJPRESS_SETTINGS["POST_PREFIX"] == "test-posts"

    # this generates a URL based on the slug only - this is prefixed with the POST_PREFIX setting
    post_url = test_post1.url

    expected_output = f'<a href="{post_url}" title="{test_post1.title}" class="u-url">{test_post1.title}</a>'
    assert djpress_tags.post_title(context) == expected_output

    # disable microformats
    settings.DJPRESS_SETTINGS["MICROFORMATS_ENABLED"] = False
    expected_output = f'<a href="{post_url}" title="{test_post1.title}">{test_post1.title}</a>'
    assert djpress_tags.post_title(context) == expected_output


@pytest.mark.django_db
def test_post_title_no_post():
    context = Context({"post": None})

    expected_output = ""
    assert djpress_tags.post_title(context) == expected_output


@pytest.mark.django_db
def test_post_title_single_post(test_post1):
    context = Context({"post": test_post1})
    expected_output = test_post1.title
    assert djpress_tags.post_title(context) == expected_output


@pytest.mark.django_db
def test_post_title_single_post_force_link(test_post1):
    context = Context({"post": test_post1})

    # this generates a URL based on the slug only - this is prefixed with the POST_PREFIX setting
    post_url = test_post1.url
    expected_output = f'<a href="{post_url}" title="{test_post1.title}" class="u-url">{test_post1.title}</a>'
    assert djpress_tags.post_title(context, force_link=True) == expected_output


@pytest.mark.django_db
def test_post_title_with_valid_tag(settings, test_post1):
    context = Context({"post": test_post1})

    # Microformats are enabled by default
    expected_output = f'<h1 class="p-name">{test_post1.title}</h1>'
    assert djpress_tags.post_title(context, outer_tag="h1") == expected_output

    # Disable microformats
    settings.DJPRESS_SETTINGS["MICROFORMATS_ENABLED"] = False

    expected_output = f"<h1>{test_post1.title}</h1>"
    assert djpress_tags.post_title(context, outer_tag="h1") == expected_output


@pytest.mark.django_db
def test_post_title_with_invalid_tag(test_post1):
    context = Context({"post": test_post1})

    expected_output = test_post1.title
    assert djpress_tags.post_title(context, outer_tag="ul") == expected_output


@pytest.mark.django_db
def test_post_title_with_prefix(settings, test_post1):
    # Confirm settings in settings_testing.py
    assert settings.DJPRESS_SETTINGS["POST_PREFIX"] == "test-posts"

    # Context should have both a posts and a post to simulate the for post in posts loop
    context = Context({"posts": [test_post1], "post": test_post1})

    post_url = test_post1.url

    expected_output = f'<a href="{post_url}" title="{test_post1.title}" class="u-url">{test_post1.title}</a>'
    assert djpress_tags.post_title(context) == expected_output


@pytest.mark.django_db
def test_get_post_author(test_post1):
    context = Context({"post": test_post1})

    author = test_post1.author
    output = get_author_display_name(author)
    assert djpress_tags.get_post_author(context) == output


def test_get_post_author_no_post():
    context = Context({"foo": "bar"})

    assert djpress_tags.get_post_author(context) == ""
    assert type(djpress_tags.get_post_author(context)) == str


def test_post_author_no_post():
    context = Context({"foo": "bar"})

    assert djpress_tags.post_author(context) == ""
    assert type(djpress_tags.post_author(context)) == str


@pytest.mark.django_db
def test_post_author(settings, test_post1):
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["AUTHOR_ENABLED"] is True
    assert settings.DJPRESS_SETTINGS["AUTHOR_PREFIX"] == "test-url-author"

    author = test_post1.author

    expected_output = (
        f'<a href="/{settings.DJPRESS_SETTINGS["AUTHOR_PREFIX"]}/testuser/" title="View all posts by '
        f'{get_author_display_name(author)}"><span class="p-author">'
        f"{get_author_display_name(author)}</span></a>"
    )
    assert djpress_tags.post_author(context) == expected_output

    # Disable microformats
    settings.DJPRESS_SETTINGS["MICROFORMATS_ENABLED"] = False
    expected_output = (
        f'<a href="/{settings.DJPRESS_SETTINGS["AUTHOR_PREFIX"]}/testuser/" title="View all posts by '
        f'{get_author_display_name(author)}"><span>'
        f"{get_author_display_name(author)}</span></a>"
    )
    assert djpress_tags.post_author(context) == expected_output


@pytest.mark.django_db
def test_post_author_author_path_disabled(settings, test_post1):
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["AUTHOR_ENABLED"] is True
    assert settings.DJPRESS_SETTINGS["AUTHOR_PREFIX"] == "test-url-author"

    settings.DJPRESS_SETTINGS["AUTHOR_ENABLED"] = False

    author = test_post1.author

    expected_output = f'<span class="p-author">{get_author_display_name(author)}</span>'
    assert djpress_tags.post_author(context) == expected_output


@pytest.mark.django_db
def test_post_author_with_author_path_with_one_link_class(settings, test_post1):
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["AUTHOR_ENABLED"] is True
    assert settings.DJPRESS_SETTINGS["AUTHOR_PREFIX"] == "test-url-author"

    author = test_post1.author

    expected_output = (
        f'<a href="/{settings.DJPRESS_SETTINGS["AUTHOR_PREFIX"]}/testuser/" title="View all posts by '
        f'{get_author_display_name(author)}" class="class1">'
        f'<span class="p-author">{get_author_display_name(author)}</span></a>'
    )
    assert djpress_tags.post_author(context, "class1") == expected_output


@pytest.mark.django_db
def test_post_author_with_author_path_with_two_link_class(settings, test_post1):
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["AUTHOR_ENABLED"] is True
    assert settings.DJPRESS_SETTINGS["AUTHOR_PREFIX"] == "test-url-author"

    author = test_post1.author

    expected_output = (
        f'<a href="/{settings.DJPRESS_SETTINGS["AUTHOR_PREFIX"]}/testuser/" title="View all posts by '
        f'{get_author_display_name(author)}" class="class1 class2">'
        f'<span class="p-author">{get_author_display_name(author)}</span></a>'
    )
    assert djpress_tags.post_author(context, "class1 class2") == expected_output


@pytest.mark.django_db
def test_post_category_link_without_category_path(settings, category1):
    """Test the post_category_link template tag without the category path enabled.

    If the CATEGORY_ENABLED setting is False, the template tag should just return
    the category name, with no link."""
    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["CATEGORY_ENABLED"] is True

    settings.DJPRESS_SETTINGS["CATEGORY_ENABLED"] = False
    assert settings.DJPRESS_SETTINGS["CATEGORY_ENABLED"] is False

    assert djpress_tags.post_category_link(category1) == category1.title


@pytest.mark.django_db
def test_post_category_link_without_category_path_with_one_link(settings, category1):
    """Test the post_category_link template tag without the category path enabled.

    If the CATEGORY_ENABLED setting is False, the template tag should just return
    the category name, with no link."""
    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["CATEGORY_ENABLED"] is True

    settings.DJPRESS_SETTINGS["CATEGORY_ENABLED"] = False
    assert settings.DJPRESS_SETTINGS["CATEGORY_ENABLED"] is False

    assert djpress_tags.post_category_link(category1, "class1") == category1.title


@pytest.mark.django_db
def test_post_category_link_with_category_path(settings, category1):
    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["CATEGORY_ENABLED"] is True
    assert settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"] == "test-url-category"

    expected_output = f'<a href="/{settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"]}/test-category1/" title="View all posts in the Test Category1 category" class="p-category">{category1.title}</a>'
    assert djpress_tags.post_category_link(category1) == expected_output

    # disable microformats
    settings.DJPRESS_SETTINGS["MICROFORMATS_ENABLED"] = False

    expected_output = f'<a href="/{settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"]}/test-category1/" title="View all posts in the Test Category1 category">{category1.title}</a>'
    assert djpress_tags.post_category_link(category1) == expected_output


@pytest.mark.django_db
def test_post_category_link_with_category_path_with_one_link_class(settings, category1):
    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["CATEGORY_ENABLED"] is True
    assert settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"] == "test-url-category"

    expected_output = f'<a href="/{settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"]}/test-category1/" title="View all posts in the Test Category1 category" class="p-category class1">{category1.title}</a>'
    assert djpress_tags.post_category_link(category1, "class1") == expected_output

    # disable microformats
    settings.DJPRESS_SETTINGS["MICROFORMATS_ENABLED"] = False

    expected_output = f'<a href="/{settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"]}/test-category1/" title="View all posts in the Test Category1 category" class="class1">{category1.title}</a>'
    assert djpress_tags.post_category_link(category1, "class1") == expected_output


@pytest.mark.django_db
def test_post_category_link_with_category_path_with_two_link_classes(settings, category1):
    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["CATEGORY_ENABLED"] is True
    assert settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"] == "test-url-category"

    expected_output = f'<a href="/{settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"]}/test-category1/" title="View all posts in the Test Category1 category" class="p-category class1 class2">{category1.title}</a>'

    assert djpress_tags.post_category_link(category1, "class1 class2") == expected_output


def test_get_post_date_no_post():
    context = Context({"foo": "bar"})

    assert djpress_tags.get_post_date(context) == ""
    assert type(djpress_tags.get_post_date(context)) == str


@pytest.mark.django_db
def test_get_post_date_with_date_archives_disabled(settings, test_post1):
    """djpress_tags.get_post_date is not impacted by the ARCHIVE_ENABLED setting."""
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["ARCHIVE_ENABLED"] is True

    settings.DJPRESS_SETTINGS["ARCHIVE_ENABLED"] = False
    assert settings.DJPRESS_SETTINGS["ARCHIVE_ENABLED"] is False

    expected_output = test_post1.date.strftime("%b %-d, %Y")

    assert djpress_tags.get_post_date(context) == expected_output


@pytest.mark.django_db
def test_get_post_date_with_date_archives_enabled(settings, test_post1):
    """djpress_tags.get_post_date is not impacted by the ARCHIVE_ENABLED setting."""
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["ARCHIVE_ENABLED"] is True

    expected_output = test_post1.date.strftime("%b %-d, %Y")

    assert djpress_tags.get_post_date(context) == expected_output


def test_post_date_no_post():
    context = Context({"foo": "bar"})

    assert djpress_tags.post_date(context) == ""
    assert type(djpress_tags.post_date(context)) == str


@pytest.mark.django_db
def test_post_date_with_date_archives_disabled(settings, test_post1):
    """Should return just the date."""
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["ARCHIVE_ENABLED"] is True

    settings.DJPRESS_SETTINGS["ARCHIVE_ENABLED"] = False
    assert settings.DJPRESS_SETTINGS["ARCHIVE_ENABLED"] is False

    expected_output = test_post1.date.strftime("%b %-d, %Y")

    assert djpress_tags.post_date(context) == expected_output


@pytest.mark.django_db
def test_post_date_with_date_archives_enabled(settings, test_post1):
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["ARCHIVE_ENABLED"] is True

    post_date = test_post1.date
    post_year = post_date.strftime("%Y")
    post_month = post_date.strftime("%m")
    post_month_name = post_date.strftime("%b")
    post_day = post_date.strftime("%d")
    post_day_name = post_date.strftime("%-d")
    post_time = post_date.strftime("%-I:%M %p")

    expected_output = (
        f'<time class="dt-published" datetime="{post_date.isoformat()}">'
        f'<a href="/test-url-archives/{post_year}/{post_month}/" title="View all posts in {post_month_name} {post_year}">{post_month_name}</a> '
        f'<a href="/test-url-archives/{post_year}/{post_month}/{post_day}/" title="View all posts on {post_day_name} {post_month_name} {post_year}">{post_day_name}</a>, '
        f'<a href="/test-url-archives/{post_year}/" title="View all posts in {post_year}">{post_year}</a>, '
        f"{post_time}."
        "</time>"
    )

    assert djpress_tags.post_date(context) == expected_output

    # disable microformats
    settings.DJPRESS_SETTINGS["MICROFORMATS_ENABLED"] = False
    expected_output = (
        f'<a href="/test-url-archives/{post_year}/{post_month}/" title="View all posts in {post_month_name} {post_year}">{post_month_name}</a> '
        f'<a href="/test-url-archives/{post_year}/{post_month}/{post_day}/" title="View all posts on {post_day_name} {post_month_name} {post_year}">{post_day_name}</a>, '
        f'<a href="/test-url-archives/{post_year}/" title="View all posts in {post_year}">{post_year}</a>, '
        f"{post_time}."
    )

    assert djpress_tags.post_date(context) == expected_output


@pytest.mark.django_db
def test_post_date_with_date_archives_enabled_with_one_link_class(settings, test_post1):
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["ARCHIVE_ENABLED"] is True

    post_date = test_post1.date
    post_year = post_date.strftime("%Y")
    post_month = post_date.strftime("%m")
    post_month_name = post_date.strftime("%b")
    post_day = post_date.strftime("%d")
    post_day_name = post_date.strftime("%-d")
    post_time = post_date.strftime("%-I:%M %p")

    expected_output = (
        f'<time class="dt-published" datetime="{post_date.isoformat()}">'
        f'<a href="/test-url-archives/{post_year}/{post_month}/" title="View all posts in {post_month_name} {post_year}" class="class1">{post_month_name}</a> '
        f'<a href="/test-url-archives/{post_year}/{post_month}/{post_day}/" title="View all posts on {post_day_name} {post_month_name} {post_year}" class="class1">{post_day_name}</a>, '
        f'<a href="/test-url-archives/{post_year}/" title="View all posts in {post_year}" class="class1">{post_year}</a>, '
        f"{post_time}."
        "</time>"
    )

    assert djpress_tags.post_date(context, "class1") == expected_output


@pytest.mark.django_db
def test_post_date_with_date_archives_enabled_with_two_link_classes(settings, test_post1):
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["ARCHIVE_ENABLED"] is True

    post_date = test_post1.date
    post_year = post_date.strftime("%Y")
    post_month = post_date.strftime("%m")
    post_month_name = post_date.strftime("%b")
    post_day = post_date.strftime("%d")
    post_day_name = post_date.strftime("%-d")
    post_time = post_date.strftime("%-I:%M %p")

    expected_output = (
        f'<time class="dt-published" datetime="{post_date.isoformat()}">'
        f'<a href="/test-url-archives/{post_year}/{post_month}/" title="View all posts in {post_month_name} {post_year}" class="class1 class2">{post_month_name}</a> '
        f'<a href="/test-url-archives/{post_year}/{post_month}/{post_day}/" title="View all posts on {post_day_name} {post_month_name} {post_year}" class="class1 class2">{post_day_name}</a>, '
        f'<a href="/test-url-archives/{post_year}/" title="View all posts in {post_year}" class="class1 class2">{post_year}</a>, '
        f"{post_time}."
        "</time>"
    )

    assert djpress_tags.post_date(context, "class1 class2") == expected_output


def test_post_content_no_post():
    """If there's no post in the context, return an empty string."""
    context = Context({"foo": "bar"})

    assert djpress_tags.post_content(context) == ""
    assert type(djpress_tags.post_content(context)) == str


@pytest.mark.django_db
def test_post_content_with_post(test_post1):
    """If there's a post in the context, return the post content."""
    context = Context({"post": test_post1})

    expected_output = f"{test_post1.content_markdown}"

    assert djpress_tags.post_content(context) == expected_output


@pytest.mark.django_db
def test_post_content_with_post_with_outer(settings, test_post1):
    """If there's a post in the context, return the post content."""
    context = Context({"post": test_post1})

    # Microformats are enabled by default
    expected_output = f'<section class="e-content">{test_post1.content_markdown}</section>'
    assert djpress_tags.post_content(context, outer_tag="section") == expected_output

    # Disable microformats
    settings.DJPRESS_SETTINGS["MICROFORMATS_ENABLED"] = False
    expected_output = f"<section>{test_post1.content_markdown}</section>"
    assert djpress_tags.post_content(context, outer_tag="section") == expected_output


@pytest.mark.django_db
def test_post_content_with_posts(test_long_post1):
    """If there's a posts in the context, return the truncated post content."""
    # Context should have both a posts and a post to simulate the for post in posts loop
    context = Context({"posts": [test_long_post1], "post": test_long_post1})

    expected_output = f"{test_long_post1.truncated_content_markdown}" f"{post_read_more_link(test_long_post1)}"

    assert djpress_tags.post_content(context) == expected_output


@pytest.mark.django_db
def test_author_name(user):
    """author_name only works if there's an author in the context."""
    context = Context({"author": user})

    # Test case 1 - no options
    expected_output = get_author_display_name(user)
    assert djpress_tags.author_name(context) == expected_output

    # Test case 2 - with options
    expected_output = f'<h1 class="title">View posts in the {get_author_display_name(user)} author</h1>'
    assert (
        djpress_tags.author_name(
            context,
            outer="h1",
            outer_class="title",
            pre_text="View posts in the ",
            post_text=" author",
        )
        == expected_output
    )


def test_author_name_no_author():
    """If there's no author in the context, return an empty string."""
    context = Context({"foo": "bar"})

    assert djpress_tags.author_name(context) == ""
    assert type(djpress_tags.author_name(context)) == str


@pytest.mark.django_db
def test_category_title(category1):
    """category_title only works if there's a category in the context."""
    context = Context({"category": category1})

    # Test case 1 - no options
    expected_output = category1.title
    assert djpress_tags.category_title(context) == expected_output

    # Test case 2 - with options
    expected_output = f'<h1 class="title">View posts in the {category1.title} category</h1>'
    assert (
        djpress_tags.category_title(
            context,
            outer="h1",
            outer_class="title",
            pre_text="View posts in the ",
            post_text=" category",
        )
        == expected_output
    )


def test_category_title_no_category():
    """If there's no category in the context, return an empty string."""
    context = Context({"foo": "bar"})

    assert djpress_tags.category_title(context) == ""
    assert type(djpress_tags.category_title(context)) == str


@pytest.mark.django_db
def test_post_categories(settings, test_post1):
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"] == "test-url-category"

    expected_output = f'<ul><li><a href="/{settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"]}/test-category1/" title="View all posts in the Test Category1 category" class="p-category">Test Category1</a></li></ul>'

    assert djpress_tags.post_categories(context) == expected_output


def test_post_categories_none_post_context():
    context = Context({"post": None})

    expected_output = ""
    assert djpress_tags.post_categories(context) == expected_output


def test_post_categories_no_post_context():
    context = Context({"foo": None})

    expected_output = ""
    assert djpress_tags.post_categories(context) == expected_output


@pytest.mark.django_db
def test_post_categories_no_categories_context(test_post1):
    test_post1.categories.clear()
    context = Context({"post": test_post1})

    expected_output = ""
    assert djpress_tags.post_categories(context) == expected_output


@pytest.mark.django_db
def test_post_categories_ul(settings, test_post1):
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"] == "test-url-category"

    expected_output = f'<ul><li><a href="/{settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"]}/test-category1/" title="View all posts in the Test Category1 category" class="p-category">Test Category1</a></li></ul>'

    assert djpress_tags.post_categories(context, "ul") == expected_output


@pytest.mark.django_db
def test_post_categories_ul_class1(settings, test_post1):
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"] == "test-url-category"

    expected_output = f'<ul><li><a href="/{settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"]}/test-category1/" title="View all posts in the Test Category1 category" class="p-category class1">Test Category1</a></li></ul>'

    assert djpress_tags.post_categories(context, outer="ul", link_class="class1") == expected_output


@pytest.mark.django_db
def test_post_categories_ul_class1_class2(settings, test_post1):
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"] == "test-url-category"

    expected_output = f'<ul><li><a href="/{settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"]}/test-category1/" title="View all posts in the Test Category1 category" class="p-category class1 class2">Test Category1</a></li></ul>'

    assert djpress_tags.post_categories(context, outer="ul", link_class="class1 class2") == expected_output


@pytest.mark.django_db
def test_post_categories_div(settings, test_post1):
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"] == "test-url-category"

    expected_output = f'<div><a href="/{settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"]}/test-category1/" title="View all posts in the Test Category1 category" class="p-category">Test Category1</a></div>'

    assert djpress_tags.post_categories(context, outer="div") == expected_output


@pytest.mark.django_db
def test_post_categories_div_class1(settings, test_post1):
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"] == "test-url-category"

    expected_output = f'<div><a href="/{settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"]}/test-category1/" title="View all posts in the Test Category1 category" class="p-category class1">Test Category1</a></div>'

    assert djpress_tags.post_categories(context, outer="div", link_class="class1") == expected_output


@pytest.mark.django_db
def test_post_categories_div_class1_class2(settings, test_post1):
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"] == "test-url-category"

    expected_output = f'<div><a href="/{settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"]}/test-category1/" title="View all posts in the Test Category1 category" class="p-category class1 class2">Test Category1</a></div>'

    assert djpress_tags.post_categories(context, outer="div", link_class="class1 class2") == expected_output


@pytest.mark.django_db
def test_post_categories_span(settings, test_post1):
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"] == "test-url-category"

    expected_output = f'<span><a href="/{settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"]}/test-category1/" title="View all posts in the Test Category1 category" class="p-category">Test Category1</a></span>'

    assert djpress_tags.post_categories(context, outer="span") == expected_output


@pytest.mark.django_db
def test_post_categories_span_class1(settings, test_post1):
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"] == "test-url-category"

    expected_output = f'<span><a href="/{settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"]}/test-category1/" title="View all posts in the Test Category1 category" class="p-category class1">Test Category1</a></span>'

    assert djpress_tags.post_categories(context, outer="span", link_class="class1") == expected_output


@pytest.mark.django_db
def test_post_categories_span_class1_class2(settings, test_post1):
    context = Context({"post": test_post1})

    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"] == "test-url-category"

    expected_output = f'<span><a href="/{settings.DJPRESS_SETTINGS["CATEGORY_PREFIX"]}/test-category1/" title="View all posts in the Test Category1 category" class="p-category class1 class2">Test Category1</a></span>'

    assert djpress_tags.post_categories(context, outer="span", link_class="class1 class2") == expected_output


@pytest.mark.django_db
def test_blog_categories(category1, category2):
    categories = Category.objects.all()

    assert category1 in categories
    assert category2 in categories

    assert djpress_tags.blog_categories() == categories_html(
        categories=categories, outer="ul", outer_class="", link_class=""
    )


@pytest.mark.django_db
def test_blog_categories_no_categories():
    assert djpress_tags.blog_categories() == ""


@pytest.mark.django_db
def test_blog_pages_list_no_pages():
    assert djpress_tags.blog_pages_list() == ""


@pytest.mark.django_db
def test_blog_pages_list_no_children(test_page1, test_page2, test_page3):
    expected_output = (
        "<ul>"
        f"<li>{get_page_link(page=test_page1)}</li>"
        f"<li>{get_page_link(page=test_page2)}</li>"
        f"<li>{get_page_link(page=test_page3)}</li>"
        "</ul>"
    )

    assert djpress_tags.blog_pages_list() == expected_output


@pytest.mark.django_db
def test_blog_pages_list_no_children_with_classes(test_page1, test_page2, test_page3):
    expected_output = (
        '<ul class="ul-outer-class">'
        f'<li class="li-class">{get_page_link(page=test_page1, link_class="a-class")}</li>'
        f'<li class="li-class">{get_page_link(page=test_page2, link_class="a-class")}</li>'
        f'<li class="li-class">{get_page_link(page=test_page3, link_class="a-class")}</li>'
        "</ul>"
    )

    assert (
        djpress_tags.blog_pages_list(
            ul_outer_class="ul-outer-class", li_class="li-class", a_class="a-class", ul_child_class="ul-child-class"
        )
        == expected_output
    )


@pytest.mark.django_db
def test_blog_pages_list_one_child(test_page1, test_page2, test_page3):
    test_page2.parent = test_page1
    test_page2.save()

    expected_output = (
        "<ul>"
        f"<li>{get_page_link(page=test_page1)}"
        "<ul>"
        f"<li>{get_page_link(page=test_page2)}</li>"
        "</ul>"
        "</li>"
        f"<li>{get_page_link(page=test_page3)}</li>"
        "</ul>"
    )

    assert djpress_tags.blog_pages_list() == expected_output


@pytest.mark.django_db
def test_blog_pages_list_one_child_with_classes(test_page1, test_page2, test_page3):
    test_page2.parent = test_page1
    test_page2.save()

    expected_output = (
        '<ul class="ul-outer-class">'
        f'<li class="li-class">{get_page_link(page=test_page1, link_class="a-class")}'
        '<ul class="ul-child-class">'
        f'<li class="li-class">{get_page_link(page=test_page2, link_class="a-class")}</li>'
        "</ul>"
        "</li>"
        f'<li class="li-class">{get_page_link(page=test_page3, link_class="a-class")}</li>'
        "</ul>"
    )

    assert (
        djpress_tags.blog_pages_list(
            ul_outer_class="ul-outer-class", li_class="li-class", a_class="a-class", ul_child_class="ul-child-class"
        )
        == expected_output
    )


@pytest.mark.django_db
def test_blog_pages_list_two_children(test_page1, test_page2, test_page3, test_page4):
    test_page3.parent = test_page1
    test_page3.save()
    test_page4.parent = test_page2
    test_page4.save()

    expected_output = (
        "<ul>"
        f"<li>{get_page_link(page=test_page1)}"
        "<ul>"
        f"<li>{get_page_link(page=test_page3)}</li>"
        "</ul>"
        "</li>"
        f"<li>{get_page_link(page=test_page2)}"
        "<ul>"
        f"<li>{get_page_link(page=test_page4)}</li>"
        "</ul>"
        "</li>"
        "</ul>"
    )

    assert djpress_tags.blog_pages_list() == expected_output


@pytest.mark.django_db
def test_blog_pages_list_two_children_with_classes(test_page1, test_page2, test_page3, test_page4):
    test_page3.parent = test_page1
    test_page3.save()
    test_page4.parent = test_page2
    test_page4.save()

    expected_output = (
        '<ul class="ul-outer-class">'
        f'<li class="li-class">{get_page_link(page=test_page1, link_class="a-class")}'
        '<ul class="ul-child-class">'
        f'<li class="li-class">{get_page_link(page=test_page3, link_class="a-class")}</li>'
        "</ul>"
        "</li>"
        f'<li class="li-class">{get_page_link(page=test_page2, link_class="a-class")}'
        '<ul class="ul-child-class">'
        f'<li class="li-class">{get_page_link(page=test_page4, link_class="a-class")}</li>'
        "</ul>"
        "</li>"
        "</ul>"
    )

    assert (
        djpress_tags.blog_pages_list(
            ul_outer_class="ul-outer-class", li_class="li-class", a_class="a-class", ul_child_class="ul-child-class"
        )
        == expected_output
    )


@pytest.mark.django_db
def test_blog_pages_list_child_grandchild(test_page1, test_page2, test_page3, test_page4):
    test_page2.parent = test_page1
    test_page2.save()
    test_page3.parent = test_page2
    test_page3.save()

    expected_output = (
        "<ul>"
        f"<li>{get_page_link(page=test_page1)}"
        "<ul>"
        f"<li>{get_page_link(page=test_page2)}"
        "<ul>"
        f"<li>{get_page_link(page=test_page3)}</li>"
        "</ul>"
        "</li>"
        "</ul>"
        "</li>"
        f"<li>{get_page_link(page=test_page4)}</li>"
        "</ul>"
    )

    assert djpress_tags.blog_pages_list() == expected_output


@pytest.mark.django_db
def test_blog_pages_list_child_grandchild_with_classes(test_page1, test_page2, test_page3, test_page4):
    test_page2.parent = test_page1
    test_page2.save()
    test_page3.parent = test_page2
    test_page3.save()

    expected_output = (
        '<ul class="ul-outer-class">'
        f'<li class="li-class">{get_page_link(page=test_page1, link_class="a-class")}'
        '<ul class="ul-child-class">'
        f'<li class="li-class">{get_page_link(page=test_page2, link_class="a-class")}'
        '<ul class="ul-child-class">'
        f'<li class="li-class">{get_page_link(page=test_page3, link_class="a-class")}</li>'
        "</ul>"
        "</li>"
        "</ul>"
        "</li>"
        f'<li class="li-class">{get_page_link(page=test_page4, link_class="a-class")}</li>'
        "</ul>"
    )

    assert (
        djpress_tags.blog_pages_list(
            ul_outer_class="ul-outer-class", li_class="li-class", a_class="a-class", ul_child_class="ul-child-class"
        )
        == expected_output
    )


@pytest.mark.django_db
def test_blog_pages_list_child_greatgrandchild(test_page1, test_page2, test_page3, test_page4):
    test_page2.parent = test_page1
    test_page2.save()
    test_page3.parent = test_page2
    test_page3.save()
    test_page4.parent = test_page3
    test_page4.save()

    expected_output = (
        "<ul>"
        f"<li>{get_page_link(page=test_page1)}"
        "<ul>"
        f"<li>{get_page_link(page=test_page2)}"
        "<ul>"
        f"<li>{get_page_link(page=test_page3)}"
        "<ul>"
        f"<li>{get_page_link(page=test_page4)}</li>"
        "</ul>"
        "</li>"
        "</ul>"
        "</li>"
        "</ul>"
        "</li>"
        "</ul>"
    )

    assert djpress_tags.blog_pages_list() == expected_output


@pytest.mark.django_db
def test_blog_pages_list_child_greatgrandchild_with_classes(test_page1, test_page2, test_page3, test_page4):
    test_page2.parent = test_page1
    test_page2.save()
    test_page3.parent = test_page2
    test_page3.save()
    test_page4.parent = test_page3
    test_page4.save()

    expected_output = (
        '<ul class="ul-outer-class">'
        f'<li class="li-class">{get_page_link(page=test_page1, link_class="a-class")}'
        '<ul class="ul-child-class">'
        f'<li class="li-class">{get_page_link(page=test_page2, link_class="a-class")}'
        '<ul class="ul-child-class">'
        f'<li class="li-class">{get_page_link(page=test_page3, link_class="a-class")}'
        '<ul class="ul-child-class">'
        f'<li class="li-class">{get_page_link(page=test_page4, link_class="a-class")}</li>'
        "</ul>"
        "</li>"
        "</ul>"
        "</li>"
        "</ul>"
        "</li>"
        "</ul>"
    )

    assert (
        djpress_tags.blog_pages_list(
            ul_outer_class="ul-outer-class", li_class="li-class", a_class="a-class", ul_child_class="ul-child-class"
        )
        == expected_output
    )


@pytest.mark.django_db
def test_blog_pages_list_child_change_order(test_page1, test_page2, test_page3, test_page4, test_page5):
    test_page5.menu_order = 1
    test_page5.save()
    test_page1.menu_order = 2
    test_page1.save()
    test_page2.parent = test_page1
    test_page2.save()
    test_page3.parent = test_page2
    test_page3.save()
    test_page4.parent = test_page1
    test_page4.save()

    expected_output = (
        "<ul>"
        f"<li>{get_page_link(page=test_page5)}</li>"
        f"<li>{get_page_link(page=test_page1)}"
        "<ul>"
        f"<li>{get_page_link(page=test_page2)}"
        "<ul>"
        f"<li>{get_page_link(page=test_page3)}</li>"
        "</ul>"
        "</li>"
        f"<li>{get_page_link(page=test_page4)}</li>"
        "</ul>"
        "</li>"
        "</ul>"
    )

    assert djpress_tags.blog_pages_list() == expected_output


@pytest.mark.django_db
def test_blog_pages_list_child_greatgreatgrandchild(test_page1, test_page2, test_page3, test_page4, test_page5):
    test_page2.parent = test_page1
    test_page2.save()
    test_page3.parent = test_page2
    test_page3.save()
    test_page4.parent = test_page3
    test_page4.save()
    test_page5.parent = test_page4
    test_page5.save()

    expected_output = (
        "<ul>"
        f"<li>{get_page_link(page=test_page1)}"
        "<ul>"
        f"<li>{get_page_link(page=test_page2)}"
        "<ul>"
        f"<li>{get_page_link(page=test_page3)}"
        "<ul>"
        f"<li>{get_page_link(page=test_page4)}"
        "<ul>"
        f"<li>{get_page_link(page=test_page5)}</li>"
        "</ul>"
        "</li>"
        "</ul>"
        "</li>"
        "</ul>"
        "</li>"
        "</ul>"
        "</li>"
        "</ul>"
    )

    assert djpress_tags.blog_pages_list() == expected_output


@pytest.mark.django_db
def test_blog_pages_no_pages():
    assert djpress_tags.blog_pages() == ""


@pytest.mark.django_db
def test_blog_pages(test_page1, test_page2):
    pages = Post.page_objects.all()

    assert test_page1 in pages
    assert test_page2 in pages

    expected_output_ul = (
        f"<ul><li>{get_page_link(page=test_page1)}</li>" f"<li>{get_page_link(page=test_page2)}</li></ul>"
    )

    expected_output_div = f"<div>{get_page_link(page=test_page1)}, {get_page_link(page=test_page2)}</div>"

    expected_output_span = f"<span>{get_page_link(page=test_page1)}, {get_page_link(page=test_page2)}</span>"

    assert djpress_tags.blog_pages() == expected_output_ul
    assert djpress_tags.blog_pages(outer="div") == expected_output_div
    assert djpress_tags.blog_pages(outer="span") == expected_output_span


@pytest.mark.django_db
def test_blog_page_title(test_post1, test_page1):
    # Test case 1 - category page
    context = Context({"category": test_post1.categories.first()})
    assert test_post1.categories.first().title == djpress_tags.blog_page_title(context)

    # Test case 2 - author page
    context = Context({"author": test_post1.author})
    assert get_author_display_name(test_post1.author) == djpress_tags.blog_page_title(context)

    # Test case 3 - single post
    context = Context({"post": test_post1})
    assert test_post1.title == djpress_tags.blog_page_title(context)

    # Test case 4 - single page
    context = Context({"post": test_page1})
    assert test_page1.title == djpress_tags.blog_page_title(context)

    # Test case 5 - no context
    context = Context()
    assert "" == djpress_tags.blog_page_title(context)


@pytest.mark.django_db
def test_is_paginated(settings, test_post1, test_post2, test_long_post1):
    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] == 3

    # Test case 1 - no paginator in context
    context = Context()
    assert djpress_tags.is_paginated(context) is False

    # Test case 2 - paginator in context
    posts = Paginator(
        Post.post_objects.get_published_posts(),
        settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"],
    )
    page = posts.get_page(number=None)
    context = Context({"posts": page})
    assert djpress_tags.is_paginated(context) is True


def test_pagination_links_no_posts():
    context = Context()

    assert djpress_tags.pagination_links(context) == ""
    assert type(djpress_tags.pagination_links(context)) == str


@pytest.mark.django_db
def test_get_pagination_range(settings, test_post1, test_post2, test_long_post1):
    # Test case 1 - 1 page, i.e. 3 posts with 3 posts per page
    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] == 3
    posts = Paginator(
        Post.post_objects.get_published_posts(),
        settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"],
    )
    page = posts.get_page(number=None)
    context = Context({"posts": page})
    assert djpress_tags.get_pagination_range(context) == range(1, 2)

    # Test case 2 - 2 pages, i.e. 3 posts with 2 posts per page
    settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] = 2
    posts = Paginator(
        Post.post_objects.get_published_posts(),
        settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"],
    )
    page = posts.get_page(number=None)
    context = Context({"posts": page})
    assert settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] == 2
    assert djpress_tags.get_pagination_range(context) == range(1, 3)

    # Test case 3 - 3 pages, i.e. 3 posts with 1 posts per page
    settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] = 1
    assert settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] == 1
    posts = Paginator(
        Post.post_objects.get_published_posts(),
        settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"],
    )
    page = posts.get_page(number=None)
    context = Context({"posts": page})
    assert djpress_tags.get_pagination_range(context) == range(1, 4)


@pytest.mark.django_db
def test_get_pagination_range_no_posts(settings):
    # Test case 1 - no posts
    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] == 3
    posts = Paginator(
        Post.post_objects.get_published_posts(),
        settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"],
    )
    page = posts.get_page(number=None)
    context = Context({"posts": page})
    assert djpress_tags.get_pagination_range(context) == range(0)


@pytest.mark.django_db
def test_get_pagination_current_page(settings, test_post1, test_post2, test_long_post1):
    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] == 3

    settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] = 1
    assert settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] == 1

    # Test case 1 - no page number
    posts = Paginator(
        Post.post_objects.get_published_posts(),
        settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"],
    )
    page = posts.get_page(number=None)
    context = Context({"posts": page})
    assert djpress_tags.get_pagination_current_page(context) == 1

    # Test case 2 - page number 1
    page = posts.get_page(number=1)
    context = Context({"posts": page})
    assert djpress_tags.get_pagination_current_page(context) == 1

    # Test case 3 - page number 2
    page = posts.get_page(number=2)
    context = Context({"posts": page})
    assert djpress_tags.get_pagination_current_page(context) == 2

    # Test case 4 - page number 3
    page = posts.get_page(number=3)
    context = Context({"posts": page})
    assert djpress_tags.get_pagination_current_page(context) == 3


@pytest.mark.django_db
def test_get_pagination_current_page_no_posts(settings):
    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] == 3

    # Test case 1 - no posts
    posts = Paginator(
        Post.post_objects.get_published_posts(),
        settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"],
    )
    page = posts.get_page(number=None)
    context = Context({"posts": page})
    assert djpress_tags.get_pagination_current_page(context) == 0


@pytest.mark.django_db
def test_pagination_links_one_page(settings, test_post1, test_post2, test_long_post1):
    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] == 3

    posts = Paginator(
        Post.post_objects.get_published_posts(),
        settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"],
    )
    page = posts.get_page(number=None)

    context = Context({"posts": page})

    previous_output = ""
    current_output = f'<span class="current">' f"Page {page.number} of {page.paginator.num_pages}" f"</span>"
    next_output = ""

    expected_output = f'<div class="pagination">{previous_output} {current_output} {next_output}</div>'

    assert djpress_tags.pagination_links(context) == expected_output


@pytest.mark.django_db
def test_pagination_links_two_pages(settings, test_post1, test_post2, test_long_post1):
    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] == 3

    settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] = 2

    assert settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] == 2

    posts = Paginator(
        Post.post_objects.get_published_posts(),
        settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"],
    )

    # Test case 1 - first page with no page
    page = posts.get_page(number=None)

    context = Context({"posts": page})

    previous_output = ""
    current_output = f'<span class="current">' f"Page {page.number} of {page.paginator.num_pages}" f"</span>"
    next_output = (
        f'<span class="next">'
        f'<a href="?page={page.next_page_number()}">next</a> '
        f'<a href="?page={page.paginator.num_pages}">last &raquo;</a>'
        f"</span>"
    )

    expected_output = f'<div class="pagination">{previous_output} {current_output} {next_output}</div>'

    assert djpress_tags.pagination_links(context) == expected_output

    # Test case 2 - first page with page number 1
    page = posts.get_page(number=1)

    context = Context({"posts": page})

    previous_output = ""
    current_output = f'<span class="current">' f"Page {page.number} of {page.paginator.num_pages}" f"</span>"
    next_output = (
        f'<span class="next">'
        f'<a href="?page={page.next_page_number()}">next</a> '
        f'<a href="?page={page.paginator.num_pages}">last &raquo;</a>'
        f"</span>"
    )

    expected_output = f'<div class="pagination">{previous_output} {current_output} {next_output}</div>'

    assert djpress_tags.pagination_links(context) == expected_output

    # Test case 3 - second page with page number 2
    page = posts.get_page(number=2)

    context = Context({"posts": page})

    previous_output = (
        f'<span class="previous">'
        f'<a href="?page=1">&laquo; first</a> '
        f'<a href="?page={page.previous_page_number()}">previous</a>'
        f"</span>"
    )
    current_output = f'<span class="current">' f"Page {page.number} of {page.paginator.num_pages}" f"</span>"
    next_output = ""

    expected_output = f'<div class="pagination">{previous_output} {current_output} {next_output}</div>'

    assert djpress_tags.pagination_links(context) == expected_output


@pytest.mark.django_db
def test_pagination_links_three_pages(settings, test_post1, test_post2, test_long_post1):
    # Confirm settings are set according to settings_testing.py
    assert settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] == 3

    settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] = 1

    assert settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] == 1

    posts = Paginator(
        Post.post_objects.get_published_posts(),
        settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"],
    )

    # Test case 1 - first page with no page number
    page = posts.get_page(number=None)

    context = Context({"posts": page})

    previous_output = ""
    current_output = f'<span class="current">' f"Page {page.number} of {page.paginator.num_pages}" f"</span>"
    next_output = (
        f'<span class="next">'
        f'<a href="?page={page.next_page_number()}">next</a> '
        f'<a href="?page={page.paginator.num_pages}">last &raquo;</a>'
        f"</span>"
    )

    expected_output = f'<div class="pagination">{previous_output} {current_output} {next_output}</div>'

    assert djpress_tags.pagination_links(context) == expected_output

    # Test case 2 - first page with page number 1
    page = posts.get_page(number=1)

    context = Context({"posts": page})

    previous_output = ""
    current_output = f'<span class="current">' f"Page {page.number} of {page.paginator.num_pages}" f"</span>"
    next_output = (
        f'<span class="next">'
        f'<a href="?page={page.next_page_number()}">next</a> '
        f'<a href="?page={page.paginator.num_pages}">last &raquo;</a>'
        f"</span>"
    )

    expected_output = f'<div class="pagination">{previous_output} {current_output} {next_output}</div>'

    assert djpress_tags.pagination_links(context) == expected_output

    # Test case 3 - first page with page number 2
    page = posts.get_page(number=2)

    context = Context({"posts": page})

    previous_output = (
        f'<span class="previous">'
        f'<a href="?page=1">&laquo; first</a> '
        f'<a href="?page={page.previous_page_number()}">previous</a>'
        f"</span>"
    )
    current_output = f'<span class="current">' f"Page {page.number} of {page.paginator.num_pages}" f"</span>"
    next_output = (
        f'<span class="next">'
        f'<a href="?page={page.next_page_number()}">next</a> '
        f'<a href="?page={page.paginator.num_pages}">last &raquo;</a>'
        f"</span>"
    )

    expected_output = f'<div class="pagination">{previous_output} {current_output} {next_output}</div>'

    assert djpress_tags.pagination_links(context) == expected_output

    # Test case 4 - first page with page number 3
    page = posts.get_page(number=3)

    context = Context({"posts": page})

    previous_output = (
        f'<span class="previous">'
        f'<a href="?page=1">&laquo; first</a> '
        f'<a href="?page={page.previous_page_number()}">previous</a>'
        f"</span>"
    )
    current_output = f'<span class="current">' f"Page {page.number} of {page.paginator.num_pages}" f"</span>"
    next_output = ""

    expected_output = f'<div class="pagination">{previous_output} {current_output} {next_output}</div>'

    assert djpress_tags.pagination_links(context) == expected_output


@pytest.mark.django_db
def test_page_link(test_page1):
    # Test 1 - page with a non-existent page_slug
    page_slug = ""
    assert djpress_tags.page_link(page_slug=page_slug) == ""

    # Test 2 - page with a page_slug and no options
    page_slug = test_page1.slug
    outer_class = ""
    link_class = ""

    output = get_page_link(page=test_page1, link_class=link_class)
    expected_output = f"<div{outer_class}>{output}</div>"
    assert djpress_tags.page_link(page_slug=page_slug) == expected_output

    # Test 3 - page with a page_slug and div and no options
    page_slug = test_page1.slug
    outer_class = ""
    link_class = ""

    output = get_page_link(page=test_page1, link_class=link_class)
    expected_output = f"<div{outer_class}>{output}</div>"
    assert djpress_tags.page_link(page_slug=page_slug) == expected_output

    # Test 4 - page with a page_slug and div and all options
    page_slug = test_page1.slug
    outer = "div"
    outer_class = "outerclass"
    link_class = "linkclass"

    outer_class = f' class="{outer_class}"' if outer_class else ""
    output = get_page_link(page=test_page1, link_class=link_class)
    expected_output = f"<div{outer_class}>{output}</div>"
    assert (
        djpress_tags.page_link(
            page_slug=page_slug,
            outer=outer,
            outer_class=outer_class,
            link_class=link_class,
        )
        == expected_output
    )

    # Test 5 - page with a page_slug and span and all options
    page_slug = test_page1.slug
    outer = "span"
    outer_class = "outerclass"
    link_class = "linkclass"

    outer_class = f' class="{outer_class}"' if outer_class else ""
    output = get_page_link(page=test_page1, link_class=link_class)
    expected_output = f"<span{outer_class}>{output}</span>"
    assert (
        djpress_tags.page_link(
            page_slug=page_slug,
            outer=outer,
            outer_class=outer_class,
            link_class=link_class,
        )
        == expected_output
    )

    # Test 6 - page with a page_slug and li and no options
    page_slug = test_page1.slug
    outer = "li"
    outer_class = ""
    link_class = ""

    outer_class = f' class="{outer_class}"' if outer_class else ""
    output = get_page_link(page=test_page1, link_class=link_class)
    expected_output = f"<li{outer_class}>{output}</li>"
    assert (
        djpress_tags.page_link(
            page_slug=page_slug,
            outer=outer,
            outer_class=outer_class,
            link_class=link_class,
        )
        == expected_output
    )

    # Test 6 - page with a page_slug and a wrong outer and no options
    page_slug = test_page1.slug
    outer = "foobar"
    outer_class = ""
    link_class = ""

    outer_class = f' class="{outer_class}"' if outer_class else ""
    output = get_page_link(page=test_page1, link_class=link_class)
    expected_output = f"{output}"
    assert (
        djpress_tags.page_link(
            page_slug=page_slug,
            outer=outer,
            outer_class=outer_class,
            link_class=link_class,
        )
        == expected_output
    )


def test_rss_url(settings):
    settings.DJPRESS_SETTINGS["RSS_ENABLED"] = True

    settings.DJPRESS_SETTINGS["RSS_PATH"] = "test-rss"
    expected_output = "/test-rss/"
    assert settings.DJPRESS_SETTINGS["RSS_PATH"] == "test-rss"
    assert settings.DJPRESS_SETTINGS["RSS_ENABLED"] is True
    assert djpress_tags.get_rss_url() == expected_output

    settings.DJPRESS_SETTINGS["RSS_PATH"] = "foobar"
    expected_output = "/foobar/"
    assert settings.DJPRESS_SETTINGS["RSS_PATH"] == "foobar"
    assert settings.DJPRESS_SETTINGS["RSS_ENABLED"] is True
    assert djpress_tags.get_rss_url() == expected_output


def test_rss_link(settings):
    settings.DJPRESS_SETTINGS["RSS_ENABLED"] = True
    settings.DJPRESS_SETTINGS["RSS_PATH"] = "test-rss"
    expected_output = f'<link rel="alternate" type="application/rss+xml" title="Latest Posts" href="/test-rss/">'
    assert djpress_tags.rss_link() == expected_output

    settings.DJPRESS_SETTINGS["RSS_ENABLED"] = False
    expected_output = ""
    assert djpress_tags.rss_link() == expected_output


@pytest.mark.django_db
def test_get_recent_posts(settings, test_post1, test_post2, test_post3):
    assert settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] == 3
    posts = Paginator(
        Post.post_objects.get_published_posts(),
        settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"],
    )
    page = posts.get_page(number=None)
    context = Context({"posts": page})
    tag_get_recent_posts = list(djpress_tags.get_recent_posts(context))
    page_posts = list(page.object_list)
    assert tag_get_recent_posts == page_posts

    # Test case 2 - 2 pages, i.e. 3 posts with 2 posts per page
    settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] = 2
    posts = Paginator(
        Post.post_objects.get_published_posts(),
        settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"],
    )
    page = posts.get_page(number=2)
    context = Context({"posts": page})
    tag_get_recent_posts = list(djpress_tags.get_recent_posts(context))
    page_posts = list(page.object_list)
    assert tag_get_recent_posts != page_posts

    # Test case 3 - 3 pages, i.e. 3 posts with 1 posts per page
    settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"] = 1
    posts = Paginator(
        Post.post_objects.get_published_posts(),
        settings.DJPRESS_SETTINGS["RECENT_PUBLISHED_POSTS_COUNT"],
    )
    page = posts.get_page(number=3)
    context = Context({"posts": page})
    tag_get_recent_posts = list(djpress_tags.get_recent_posts(context))
    page_posts = list(page.object_list)
    assert tag_get_recent_posts != page_posts


@pytest.mark.django_db
def test_post_wrapper_single_post_no_tag(settings):
    """Creates an article by default."""
    template_text = "{% load djpress_tags %}{% post_wrap %}<p>This is test post 1.</p>{% end_post_wrap %}"
    context = Context({})

    template = Template(template_text)
    expected_output = '<article class="h-entry"><p>This is test post 1.</p></article>'
    assert template.render(context) == expected_output

    # Disable microformats
    settings.DJPRESS_SETTINGS["MICROFORMATS_ENABLED"] = False

    template = Template(template_text)
    expected_output = "<article><p>This is test post 1.</p></article>"
    assert template.render(context) == expected_output


@pytest.mark.django_db
def test_post_wrapper_single_post_with_valid_tag(settings):
    """Create a div instead."""
    template_text = "{% load djpress_tags %}{% post_wrap tag='div' %}<p>This is test post 1.</p>{% end_post_wrap %}"
    context = Context({})

    template = Template(template_text)
    expected_output = '<div class="h-entry"><p>This is test post 1.</p></div>'
    assert template.render(context) == expected_output

    # Disable microformats
    settings.DJPRESS_SETTINGS["MICROFORMATS_ENABLED"] = False

    template = Template(template_text)
    expected_output = "<div><p>This is test post 1.</p></div>"
    assert template.render(context) == expected_output


@pytest.mark.django_db
def test_post_wrapper_single_post_with_valid_tag_double_quotes(settings):
    """Create a div instead."""
    template_text = '{% load djpress_tags %}{% post_wrap tag="div" %}<p>This is test post 1.</p>{% end_post_wrap %}'
    context = Context({})

    template = Template(template_text)
    expected_output = '<div class="h-entry"><p>This is test post 1.</p></div>'
    assert template.render(context) == expected_output

    # Disable microformats
    settings.DJPRESS_SETTINGS["MICROFORMATS_ENABLED"] = False

    template = Template(template_text)
    expected_output = "<div><p>This is test post 1.</p></div>"
    assert template.render(context) == expected_output


@pytest.mark.django_db
def test_post_wrapper_single_post_with_invalid_tag():
    """Just returns the content."""
    template_text = "{% load djpress_tags %}{% post_wrap tag='foobar' %}<p>This is test post 1.</p>{% end_post_wrap %}"
    context = Context({})

    template = Template(template_text)
    expected_output = "<p>This is test post 1.</p>"
    assert template.render(context) == expected_output


@pytest.mark.django_db
def test_post_wrapper_single_post_with_valid_tag_arg_only():
    """Just returns the content."""
    template_text = "{% load djpress_tags %}{% post_wrap 'article' %}<p>This is test post 1.</p>{% end_post_wrap %}"
    context = Context({})

    template = Template(template_text)
    expected_output = '<article class="h-entry"><p>This is test post 1.</p></article>'
    assert template.render(context) == expected_output


@pytest.mark.django_db
def test_post_wrapper_single_post_with_invalid_tag_arg_only():
    """Just returns the content."""
    template_text = "{% load djpress_tags %}{% post_wrap 'foobar' %}<p>This is test post 1.</p>{% end_post_wrap %}"
    context = Context({})

    template = Template(template_text)
    expected_output = "<p>This is test post 1.</p>"
    assert template.render(context) == expected_output


@pytest.mark.django_db
def test_post_wrapper_single_post_with_class(settings):
    template_text = (
        "{% load djpress_tags %}{% post_wrap class='blog-post' %}<p>This is test post 1.</p>{% end_post_wrap %}"
    )
    context = Context({})

    template = Template(template_text)
    expected_output = '<article class="h-entry blog-post"><p>This is test post 1.</p></article>'
    assert template.render(context) == expected_output

    # Disable microformats
    settings.DJPRESS_SETTINGS["MICROFORMATS_ENABLED"] = False
    template = Template(template_text)
    expected_output = '<article class="blog-post"><p>This is test post 1.</p></article>'
    assert template.render(context) == expected_output


@pytest.mark.django_db
def test_post_wrapper_single_post_with_tag_and_class(settings):
    template_text = "{% load djpress_tags %}{% post_wrap tag='div' class='blog-post' %}<p>This is test post 1.</p>{% end_post_wrap %}"
    context = Context({})

    template = Template(template_text)
    expected_output = '<div class="h-entry blog-post"><p>This is test post 1.</p></div>'
    assert template.render(context) == expected_output

    # Disable microformats
    settings.DJPRESS_SETTINGS["MICROFORMATS_ENABLED"] = False
    template = Template(template_text)
    expected_output = '<div class="blog-post"><p>This is test post 1.</p></div>'
    assert template.render(context) == expected_output
